from typing import Any, Mapping

from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.core.rl_module.rl_module import RLModule, RLModuleConfig
from ray.rllib.models.torch.torch_distributions import TorchCategorical
from ray.rllib.core.rl_module.marl_module import (
    MultiAgentRLModuleConfig,
    MultiAgentRLModule,
)
from ray.rllib.core.rl_module.torch.torch_rl_module import TorchRLModule
from ray.rllib.core.models.specs.typing import SpecType
from ray.rllib.utils.annotations import override
from ray.rllib.utils.framework import try_import_torch
from ray.rllib.utils.nested_dict import NestedDict

torch, nn = try_import_torch()


class DiscreteBCTorchModule(TorchRLModule):
    def __init__(self, config: RLModuleConfig) -> None:
        super().__init__(config)

    def setup(self):
        input_dim = self.config.observation_space.shape[0]
        hidden_dim = self.config.model_config_dict["fcnet_hiddens"][0]
        output_dim = self.config.action_space.n

        self.policy = nn.Sequential(
            nn.Linear(input_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, output_dim),
        )

        self.input_dim = input_dim

    def get_train_action_dist_cls(self):
        return TorchCategorical

    def get_exploration_action_dist_cls(self):
        return TorchCategorical

    def get_inference_action_dist_cls(self):
        return TorchCategorical

    @override(RLModule)
    def output_specs_exploration(self) -> SpecType:
        return [SampleBatch.ACTION_DIST_INPUTS]

    @override(RLModule)
    def output_specs_inference(self) -> SpecType:
        return [SampleBatch.ACTION_DIST_INPUTS]

    @override(RLModule)
    def output_specs_train(self) -> SpecType:
        return [SampleBatch.ACTION_DIST_INPUTS]

    @override(RLModule)
    def _forward_inference(self, batch: NestedDict) -> Mapping[str, Any]:
        with torch.no_grad():
            return self._forward_train(batch)

    @override(RLModule)
    def _forward_exploration(self, batch: NestedDict) -> Mapping[str, Any]:
        with torch.no_grad():
            return self._forward_train(batch)

    @override(RLModule)
    def _forward_train(self, batch: NestedDict) -> Mapping[str, Any]:
        action_logits = self.policy(batch["obs"])
        return {SampleBatch.ACTION_DIST_INPUTS: action_logits}


class BCTorchRLModuleWithSharedGlobalEncoder(TorchRLModule):
    """An example of an RLModule that uses an encoder shared with other things.

    For example, we could consider a multi-agent case where for inference each agent
    needs to know the global state of the environment, as well as the local state of
    itself. For better representation learning we would like to share the encoder
    across all the modules. So this module simply accepts the encoder object as its
    input argument and uses it to encode the global state. The local state is passed
    through as is. The policy head is then a simple MLP that takes the concatenation of
    the global and local state as input and outputs the action logits.

    """

    def __init__(
        self, encoder: nn.Module, local_dim: int, hidden_dim: int, action_dim: int
    ) -> None:
        super().__init__(config=None)

        self.encoder = encoder
        self.policy_head = nn.Sequential(
            nn.Linear(hidden_dim + local_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, action_dim),
        )

    def get_train_action_dist_cls(self):
        return TorchCategorical

    def get_exploration_action_dist_cls(self):
        return TorchCategorical

    def get_inference_action_dist_cls(self):
        return TorchCategorical

    @override(RLModule)
    def _default_input_specs(self):
        return [("obs", "global"), ("obs", "local")]

    @override(RLModule)
    def _forward_inference(self, batch):
        with torch.no_grad():
            return self._common_forward(batch)

    @override(RLModule)
    def _forward_exploration(self, batch):
        with torch.no_grad():
            return self._common_forward(batch)

    @override(RLModule)
    def _forward_train(self, batch):
        return self._common_forward(batch)

    def _common_forward(self, batch):
        obs = batch["obs"]
        global_enc = self.encoder(obs["global"])
        policy_in = torch.cat([global_enc, obs["local"]], dim=-1)
        action_logits = self.policy_head(policy_in)

        return {SampleBatch.ACTION_DIST_INPUTS: action_logits}


class BCTorchMultiAgentModuleWithSharedEncoder(MultiAgentRLModule):
    def __init__(self, config: MultiAgentRLModuleConfig) -> None:
        super().__init__(config)

    def setup(self):
        module_specs = self.config.modules
        module_spec = next(iter(module_specs.values()))
        global_dim = module_spec.observation_space["global"].shape[0]
        hidden_dim = module_spec.model_config_dict["fcnet_hiddens"][0]
        shared_encoder = nn.Sequential(
            nn.Linear(global_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim),
        )

        rl_modules = {}
        for module_id, module_spec in module_specs.items():
            rl_modules[module_id] = module_spec.module_class(
                encoder=shared_encoder,
                local_dim=module_spec.observation_space["local"].shape[0],
                hidden_dim=hidden_dim,
                action_dim=module_spec.action_space.n,
            )

        self._rl_modules = rl_modules

    def serialize(self):
        # TODO (Kourosh): Implement when needed.
        raise NotImplementedError

    def deserialize(self, data):
        # TODO (Kourosh): Implement when needed.
        raise NotImplementedError



import uuid
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from ray.util.annotations import PublicAPI

if TYPE_CHECKING:
    from ray.runtime_env import RuntimeEnv


@PublicAPI
class JobConfig:
    """A class used to store the configurations of a job.

    Examples:
        .. testcode::
            :hide:

            import ray
            ray.shutdown()

        .. testcode::

            import ray
            from ray.job_config import JobConfig

            ray.init(job_config=JobConfig(default_actor_lifetime="non_detached"))

    Args:
        jvm_options: The jvm options for java workers of the job.
        code_search_path: A list of directories or jar files that
            specify the search path for user code. This will be used as
            `CLASSPATH` in Java and `PYTHONPATH` in Python.
            See :ref:`Ray cross-language programming <cross_language>` for more details.
        runtime_env: A :ref:`runtime environment <runtime-environments>` dictionary.
        metadata: An opaque metadata dictionary.
        ray_namespace: A :ref:`namespace <namespaces-guide>`
            is a logical grouping of jobs and named actors.
        default_actor_lifetime: The default value of actor lifetime,
            can be "detached" or "non_detached".
            See :ref:`actor lifetimes <actor-lifetimes>` for more details.
    """

    def __init__(
        self,
        jvm_options: Optional[List[str]] = None,
        code_search_path: Optional[List[str]] = None,
        runtime_env: Optional[dict] = None,
        _client_job: bool = False,
        metadata: Optional[dict] = None,
        ray_namespace: Optional[str] = None,
        default_actor_lifetime: str = "non_detached",
        _py_driver_sys_path: Optional[List[str]] = None,
    ):
        #: The jvm options for java workers of the job.
        self.jvm_options = jvm_options or []
        #: A list of directories or jar files that
        #: specify the search path for user code.
        self.code_search_path = code_search_path or []
        # It's difficult to find the error that caused by the
        # code_search_path is a string. So we assert here.
        assert isinstance(self.code_search_path, (list, tuple)), (
            f"The type of code search path is incorrect: " f"{type(code_search_path)}"
        )
        self._client_job = _client_job
        #: An opaque metadata dictionary.
        self.metadata = metadata or {}
        #: A namespace is a logical grouping of jobs and named actors.
        self.ray_namespace = ray_namespace
        self.set_runtime_env(runtime_env)
        self.set_default_actor_lifetime(default_actor_lifetime)
        # A list of directories that specify the search path for python workers.
        self._py_driver_sys_path = _py_driver_sys_path or []

    def set_metadata(self, key: str, value: str) -> None:
        """Add key-value pair to the metadata dictionary.

        If the key already exists, the value is overwritten to the new value.

        Examples:
            .. testcode::

                import ray
                from ray.job_config import JobConfig

                job_config = JobConfig()
                job_config.set_metadata("submitter", "foo")

        Args:
            key: The key of the metadata.
            value: The value of the metadata.
        """
        self.metadata[key] = value

    def _serialize(self) -> str:
        """Serialize the struct into protobuf string"""
        return self._get_proto_job_config().SerializeToString()

    def set_runtime_env(
        self,
        runtime_env: Optional[Union[Dict[str, Any], "RuntimeEnv"]],
        validate: bool = False,
    ) -> None:
        """Modify the runtime_env of the JobConfig.

        We don't validate the runtime_env by default here because it may go
        through some translation before actually being passed to C++ (e.g.,
        working_dir translated from a local directory to a URI).

        Args:
            runtime_env: A :ref:`runtime environment <runtime-environments>` dictionary.
            validate: Whether to validate the runtime env.
        """
        self.runtime_env = runtime_env if runtime_env is not None else {}
        if validate:
            self.runtime_env = self._validate_runtime_env()
        self._cached_pb = None

    def set_ray_namespace(self, ray_namespace: str) -> None:
        """Set Ray :ref:`namespace <namespaces-guide>`.

        Args:
            ray_namespace: The namespace to set.
        """

        if ray_namespace != self.ray_namespace:
            self.ray_namespace = ray_namespace
            self._cached_pb = None

    def set_default_actor_lifetime(self, default_actor_lifetime: str) -> None:
        """Set the default actor lifetime, which can be "detached" or "non_detached".

        See :ref:`actor lifetimes <actor-lifetimes>` for more details.

        Args:
            default_actor_lifetime: The default actor lifetime to set.
        """
        import ray.core.generated.common_pb2 as common_pb2

        if default_actor_lifetime == "detached":
            self._default_actor_lifetime = common_pb2.JobConfig.ActorLifetime.DETACHED
        elif default_actor_lifetime == "non_detached":
            self._default_actor_lifetime = (
                common_pb2.JobConfig.ActorLifetime.NON_DETACHED
            )
        else:
            raise ValueError(
                "Default actor lifetime must be one of `detached`, `non_detached`"
            )

    def _validate_runtime_env(self):
        # TODO(edoakes): this is really unfortunate, but JobConfig is imported
        # all over the place so this causes circular imports. We should remove
        # this dependency and pass in a validated runtime_env instead.
        from ray.runtime_env import RuntimeEnv

        if isinstance(self.runtime_env, RuntimeEnv):
            return self.runtime_env
        return RuntimeEnv(**self.runtime_env)

    def _get_proto_job_config(self):
        """Return the protobuf structure of JobConfig."""
        # TODO(edoakes): this is really unfortunate, but JobConfig is imported
        # all over the place so this causes circular imports. We should remove
        # this dependency and pass in a validated runtime_env instead.
        import ray.core.generated.common_pb2 as common_pb2
        from ray._private.utils import get_runtime_env_info

        if self._cached_pb is None:
            pb = common_pb2.JobConfig()
            if self.ray_namespace is None:
                pb.ray_namespace = str(uuid.uuid4())
            else:
                pb.ray_namespace = self.ray_namespace
            pb.jvm_options.extend(self.jvm_options)
            pb.code_search_path.extend(self.code_search_path)
            pb.py_driver_sys_path.extend(self._py_driver_sys_path)
            for k, v in self.metadata.items():
                pb.metadata[k] = v

            parsed_env = self._validate_runtime_env()
            pb.runtime_env_info.CopyFrom(
                get_runtime_env_info(
                    parsed_env,
                    is_job_runtime_env=True,
                    serialize=False,
                )
            )

            if self._default_actor_lifetime is not None:
                pb.default_actor_lifetime = self._default_actor_lifetime
            self._cached_pb = pb

        return self._cached_pb

    def _runtime_env_has_working_dir(self):
        return self._validate_runtime_env().has_working_dir()

    def _get_serialized_runtime_env(self) -> str:
        """Return the JSON-serialized parsed runtime env dict"""
        return self._validate_runtime_env().serialize()

    def _get_proto_runtime_env_config(self) -> str:
        """Return the JSON-serialized parsed runtime env info"""
        return self._get_proto_job_config().runtime_env_info.runtime_env_config

    @classmethod
    def from_json(cls, job_config_json):
        """Generates a JobConfig object from json.

        Examples:
            .. testcode::

                from ray.job_config import JobConfig

                job_config = JobConfig.from_json(
                    {"runtime_env": {"working_dir": "uri://abc"}})

        Args:
            job_config_json: The job config json dictionary.
        """
        return cls(
            jvm_options=job_config_json.get("jvm_options", None),
            code_search_path=job_config_json.get("code_search_path", None),
            runtime_env=job_config_json.get("runtime_env", None),
            metadata=job_config_json.get("metadata", None),
            ray_namespace=job_config_json.get("ray_namespace", None),
            _client_job=job_config_json.get("client_job", False),
            _py_driver_sys_path=job_config_json.get("py_driver_sys_path", None),
        )



import logging
from typing import Any, Tuple, TYPE_CHECKING

from ray.rllib.connectors.action.clip import ClipActionsConnector
from ray.rllib.connectors.action.immutable import ImmutableActionsConnector
from ray.rllib.connectors.action.lambdas import ConvertToNumpyConnector
from ray.rllib.connectors.action.normalize import NormalizeActionsConnector
from ray.rllib.connectors.action.pipeline import ActionConnectorPipeline
from ray.rllib.connectors.agent.clip_reward import ClipRewardAgentConnector
from ray.rllib.connectors.agent.obs_preproc import ObsPreprocessorConnector
from ray.rllib.connectors.agent.pipeline import AgentConnectorPipeline
from ray.rllib.connectors.agent.state_buffer import StateBufferConnector
from ray.rllib.connectors.agent.view_requirement import ViewRequirementAgentConnector
from ray.rllib.connectors.connector import Connector, ConnectorContext
from ray.rllib.connectors.registry import get_connector
from ray.rllib.connectors.agent.mean_std_filter import (
    MeanStdObservationFilterAgentConnector,
    ConcurrentMeanStdObservationFilterAgentConnector,
)
from ray.util.annotations import PublicAPI, DeveloperAPI
from ray.rllib.connectors.agent.synced_filter import SyncedFilterAgentConnector

if TYPE_CHECKING:
    from ray.rllib.algorithms.algorithm_config import AlgorithmConfig
    from ray.rllib.policy.policy import Policy

logger = logging.getLogger(__name__)


def __preprocessing_enabled(config: "AlgorithmConfig"):
    if config._disable_preprocessor_api:
        return False
    # Same conditions as in RolloutWorker.__init__.
    if config.is_atari and config.preprocessor_pref == "deepmind":
        return False
    if config.preprocessor_pref is None:
        return False
    return True


def __clip_rewards(config: "AlgorithmConfig"):
    # Same logic as in RolloutWorker.__init__.
    # We always clip rewards for Atari games.
    return config.clip_rewards or config.is_atari


@PublicAPI(stability="alpha")
def get_agent_connectors_from_config(
    ctx: ConnectorContext,
    config: "AlgorithmConfig",
) -> AgentConnectorPipeline:
    connectors = []

    clip_rewards = __clip_rewards(config)
    if clip_rewards is True:
        connectors.append(ClipRewardAgentConnector(ctx, sign=True))
    elif type(clip_rewards) == float:
        connectors.append(ClipRewardAgentConnector(ctx, limit=abs(clip_rewards)))

    if __preprocessing_enabled(config):
        connectors.append(ObsPreprocessorConnector(ctx))

    # Filters should be after observation preprocessing
    filter_connector = get_synced_filter_connector(
        ctx,
    )
    # Configuration option "NoFilter" results in `filter_connector==None`.
    if filter_connector:
        connectors.append(filter_connector)

    connectors.extend(
        [
            StateBufferConnector(ctx),
            ViewRequirementAgentConnector(ctx),
        ]
    )

    return AgentConnectorPipeline(ctx, connectors)


@PublicAPI(stability="alpha")
def get_action_connectors_from_config(
    ctx: ConnectorContext,
    config: "AlgorithmConfig",
) -> ActionConnectorPipeline:
    """Default list of action connectors to use for a new policy.

    Args:
        ctx: context used to create connectors.
        config: The AlgorithmConfig object.
    """
    connectors = [ConvertToNumpyConnector(ctx)]
    if config.get("normalize_actions", False):
        connectors.append(NormalizeActionsConnector(ctx))
    if config.get("clip_actions", False):
        connectors.append(ClipActionsConnector(ctx))
    connectors.append(ImmutableActionsConnector(ctx))
    return ActionConnectorPipeline(ctx, connectors)


@PublicAPI(stability="alpha")
def create_connectors_for_policy(policy: "Policy", config: "AlgorithmConfig"):
    """Util to create agent and action connectors for a Policy.

    Args:
        policy: Policy instance.
        config: Algorithm config dict.
    """
    ctx: ConnectorContext = ConnectorContext.from_policy(policy)

    assert (
        policy.agent_connectors is None and policy.action_connectors is None
    ), "Can not create connectors for a policy that already has connectors."

    policy.agent_connectors = get_agent_connectors_from_config(ctx, config)
    policy.action_connectors = get_action_connectors_from_config(ctx, config)

    logger.info("Using connectors:")
    logger.info(policy.agent_connectors.__str__(indentation=4))
    logger.info(policy.action_connectors.__str__(indentation=4))


@PublicAPI(stability="alpha")
def restore_connectors_for_policy(
    policy: "Policy", connector_config: Tuple[str, Tuple[Any]]
) -> Connector:
    """Util to create connector for a Policy based on serialized config.

    Args:
        policy: Policy instance.
        connector_config: Serialized connector config.
    """
    ctx: ConnectorContext = ConnectorContext.from_policy(policy)
    name, params = connector_config
    return get_connector(name, ctx, params)


# We need this filter selection mechanism temporarily to remain compatible to old API
@DeveloperAPI
def get_synced_filter_connector(ctx: ConnectorContext):
    filter_specifier = ctx.config.get("observation_filter")
    if filter_specifier == "MeanStdFilter":
        return MeanStdObservationFilterAgentConnector(ctx, clip=None)
    elif filter_specifier == "ConcurrentMeanStdFilter":
        return ConcurrentMeanStdObservationFilterAgentConnector(ctx, clip=None)
    elif filter_specifier == "NoFilter":
        return None
    else:
        raise Exception("Unknown observation_filter: " + str(filter_specifier))


@DeveloperAPI
def maybe_get_filters_for_syncing(rollout_worker, policy_id):
    # As long as the historic filter synchronization mechanism is in
    # place, we need to put filters into self.filters so that they get
    # synchronized
    policy = rollout_worker.policy_map[policy_id]
    if not policy.agent_connectors:
        return

    filter_connectors = policy.agent_connectors[SyncedFilterAgentConnector]
    # There can only be one filter at a time
    if not filter_connectors:
        return

    assert len(filter_connectors) == 1, (
        "ConnectorPipeline has multiple connectors of type "
        "SyncedFilterAgentConnector but can only have one."
    )
    rollout_worker.filters[policy_id] = filter_connectors[0].filter



"""
This file implements a MobileNet v2 Encoder.
It uses MobileNet v2 to encode images into a latent space of 1000 dimensions.

Depending on the experiment, the MobileNet v2 encoder layers can be frozen or
unfrozen. This is controlled by the `freeze` parameter in the config.

This is an example of how a pre-trained neural network can be used as an encoder
in RLlib. You can modify this example to accommodate your own encoder network or
other pre-trained networks.
"""

from ray.rllib.core.models.base import Encoder, ENCODER_OUT
from ray.rllib.core.models.configs import ModelConfig
from ray.rllib.core.models.torch.base import TorchModel
from ray.rllib.utils.framework import try_import_torch

torch, nn = try_import_torch()

MOBILENET_INPUT_SHAPE = (3, 224, 224)


class MobileNetV2EncoderConfig(ModelConfig):
    # MobileNet v2 has a flat output with a length of 1000.
    output_dims = (1000,)
    freeze = True

    def build(self, framework):
        assert framework == "torch", "Unsupported framework `{}`!".format(framework)
        return MobileNetV2Encoder(self)


class MobileNetV2Encoder(TorchModel, Encoder):
    """A MobileNet v2 encoder for RLlib."""

    def __init__(self, config):
        super().__init__(config)
        self.net = torch.hub.load(
            "pytorch/vision:v0.6.0", "mobilenet_v2", pretrained=True
        )
        if config.freeze:
            # We don't want to train this encoder, so freeze its parameters!
            for p in self.net.parameters():
                p.requires_grad = False

    def _forward(self, input_dict, **kwargs):
        return {ENCODER_OUT: (self.net(input_dict["obs"]))}



from typing import Union

from ray.rllib.models.action_dist import ActionDistribution
from ray.rllib.utils.annotations import override
from ray.rllib.utils.deprecation import Deprecated
from ray.rllib.utils.exploration.exploration import TensorType
from ray.rllib.utils.exploration.soft_q import SoftQ
from ray.rllib.utils.framework import try_import_tf, try_import_torch

tf1, tf, tfv = try_import_tf()
torch, _ = try_import_torch()


@Deprecated(error=False)
class SlateSoftQ(SoftQ):
    @override(SoftQ)
    def get_exploration_action(
        self,
        action_distribution: ActionDistribution,
        timestep: Union[int, TensorType],
        explore: bool = True,
    ):
        assert (
            self.framework == "torch"
        ), "ERROR: SlateSoftQ only supports torch so far!"

        cls = type(action_distribution)

        # Re-create the action distribution with the correct temperature
        # applied.
        action_distribution = cls(
            action_distribution.inputs, self.model, temperature=self.temperature
        )
        batch_size = action_distribution.inputs.size()[0]
        action_logp = torch.zeros(batch_size, dtype=torch.float)

        self.last_timestep = timestep

        # Explore.
        if explore:
            # Return stochastic sample over (q-value) logits.
            action = action_distribution.sample()
        # Return the deterministic "sample" (argmax) over (q-value) logits.
        else:
            action = action_distribution.deterministic_sample()

        return action, action_logp



"""
This file stores global state for a Serve application. Deployment replicas
can use this state to access metadata or the Serve controller.
"""

import logging
from dataclasses import dataclass
from typing import Callable

import ray
from ray.exceptions import RayActorError
from ray.serve._private.client import ServeControllerClient
from ray.serve._private.common import ReplicaTag
from ray.serve._private.constants import SERVE_CONTROLLER_NAME, SERVE_NAMESPACE
from ray.serve.exceptions import RayServeException
from ray.util.annotations import PublicAPI, DeveloperAPI
import contextvars

logger = logging.getLogger(__file__)

_INTERNAL_REPLICA_CONTEXT: "ReplicaContext" = None
_global_client: ServeControllerClient = None


@PublicAPI(stability="alpha")
@dataclass
class ReplicaContext:
    """Stores data for Serve API calls from within deployments."""

    deployment: str
    replica_tag: ReplicaTag
    _internal_controller_name: str
    servable_object: Callable
    app_name: str


@PublicAPI(stability="alpha")
def get_global_client(_health_check_controller: bool = False) -> ServeControllerClient:
    """Gets the global client, which stores the controller's handle.

    Args:
        _health_check_controller: If True, run a health check on the
            cached controller if it exists. If the check fails, try reconnecting
            to the controller.

    Raises:
        RayServeException: if there is no running Serve controller actor.
    """

    try:
        if _global_client is not None:
            if _health_check_controller:
                ray.get(_global_client._controller.check_alive.remote())
            return _global_client
    except RayActorError:
        logger.info("The cached controller has died. Reconnecting.")
        _set_global_client(None)

    return _connect()


def _set_global_client(client):
    global _global_client
    _global_client = client


@PublicAPI(stability="alpha")
def get_internal_replica_context():
    return _INTERNAL_REPLICA_CONTEXT


def _set_internal_replica_context(
    deployment: str,
    replica_tag: ReplicaTag,
    controller_name: str,
    servable_object: Callable,
    app_name: str,
):
    global _INTERNAL_REPLICA_CONTEXT
    _INTERNAL_REPLICA_CONTEXT = ReplicaContext(
        deployment, replica_tag, controller_name, servable_object, app_name
    )


def _connect() -> ServeControllerClient:
    """Connect to an existing Serve application on this Ray cluster.

    If calling from the driver program, the Serve app on this Ray cluster
    must first have been initialized using `serve.start(detached=True)`.

    If called from within a replica, this will connect to the same Serve
    app that the replica is running in.

    Returns:
        ServeControllerClient that encapsulates a Ray actor handle to the
        existing Serve application's Serve Controller.

    Raises:
        RayServeException: if there is no running Serve controller actor.
    """

    # Initialize ray if needed.
    ray._private.worker.global_worker._filter_logs_by_job = False
    if not ray.is_initialized():
        ray.init(namespace=SERVE_NAMESPACE)

    # When running inside of a replica, _INTERNAL_REPLICA_CONTEXT is set to
    # ensure that the correct instance is connected to.
    if _INTERNAL_REPLICA_CONTEXT is None:
        controller_name = SERVE_CONTROLLER_NAME
    else:
        controller_name = _INTERNAL_REPLICA_CONTEXT._internal_controller_name

    # Try to get serve controller if it exists
    try:
        controller = ray.get_actor(controller_name, namespace=SERVE_NAMESPACE)
    except ValueError:
        raise RayServeException(
            "There is no "
            "instance running on this Ray cluster. Please "
            "call `serve.start(detached=True) to start "
            "one."
        )

    client = ServeControllerClient(
        controller,
        controller_name,
        detached=True,
    )
    _set_global_client(client)
    return client


# Serve request context var which is used for storing the internal
# request context information.
# route_prefix: http url route path, e.g. http://127.0.0.1:/app
#     the route is "/app". When you send requests by handle,
#     the route is empty.
# request_id: the request id is generated from http proxy, the value
#     shouldn't be changed when the variable is set.
# note:
#   The request context is readonly to avoid potential
#       async task conflicts when using it concurrently.


@DeveloperAPI
@dataclass(frozen=True)
class RequestContext:
    route: str = ""
    request_id: str = ""
    app_name: str = ""
    multiplexed_model_id: str = ""


_serve_request_context = contextvars.ContextVar(
    "Serve internal request context variable", default=RequestContext()
)


def _set_request_context(
    route: str = "",
    request_id: str = "",
    app_name: str = "",
    multiplexed_model_id: str = "",
):
    """Set the request context. If the value is not set,
    the current context value will be used."""

    current_request_context = _serve_request_context.get()
    _serve_request_context.set(
        RequestContext(
            route=route or current_request_context.route,
            request_id=request_id or current_request_context.request_id,
            app_name=app_name or current_request_context.app_name,
            multiplexed_model_id=multiplexed_model_id
            or current_request_context.multiplexed_model_id,
        )
    )



import numpy as np
from typing import Dict, List
import gymnasium as gym

from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
from ray.rllib.models.torch.misc import (
    normc_initializer,
    same_padding,
    SlimConv2d,
    SlimFC,
)
from ray.rllib.models.utils import get_activation_fn, get_filter_config
from ray.rllib.utils.annotations import override
from ray.rllib.utils.framework import try_import_torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from ray.rllib.utils.deprecation import deprecation_warning
from ray.util import log_once

torch, nn = try_import_torch()


class VisionNetwork(TorchModelV2, nn.Module):
    """Generic vision network."""

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
    ):
        if log_once("torch_visionnet_deprecation"):
            deprecation_warning(old="ray.rllib.models.torch.visionnet.VisionNetwork")

        if not model_config.get("conv_filters"):
            model_config["conv_filters"] = get_filter_config(obs_space.shape)

        TorchModelV2.__init__(
            self, obs_space, action_space, num_outputs, model_config, name
        )
        nn.Module.__init__(self)

        activation = self.model_config.get("conv_activation")
        filters = self.model_config["conv_filters"]
        assert len(filters) > 0, "Must provide at least 1 entry in `conv_filters`!"

        # Post FC net config.
        post_fcnet_hiddens = model_config.get("post_fcnet_hiddens", [])
        post_fcnet_activation = get_activation_fn(
            model_config.get("post_fcnet_activation"), framework="torch"
        )

        no_final_linear = self.model_config.get("no_final_linear")
        vf_share_layers = self.model_config.get("vf_share_layers")

        # Whether the last layer is the output of a Flattened (rather than
        # a n x (1,1) Conv2D).
        self.last_layer_is_flattened = False
        self._logits = None

        layers = []
        (w, h, in_channels) = obs_space.shape

        in_size = [w, h]
        for out_channels, kernel, stride in filters[:-1]:
            padding, out_size = same_padding(in_size, kernel, stride)
            layers.append(
                SlimConv2d(
                    in_channels,
                    out_channels,
                    kernel,
                    stride,
                    padding,
                    activation_fn=activation,
                )
            )
            in_channels = out_channels
            in_size = out_size

        out_channels, kernel, stride = filters[-1]

        # No final linear: Last layer has activation function and exits with
        # num_outputs nodes (this could be a 1x1 conv or a FC layer, depending
        # on `post_fcnet_...` settings).
        if no_final_linear and num_outputs:
            out_channels = out_channels if post_fcnet_hiddens else num_outputs
            layers.append(
                SlimConv2d(
                    in_channels,
                    out_channels,
                    kernel,
                    stride,
                    None,  # padding=valid
                    activation_fn=activation,
                )
            )

            # Add (optional) post-fc-stack after last Conv2D layer.
            layer_sizes = post_fcnet_hiddens[:-1] + (
                [num_outputs] if post_fcnet_hiddens else []
            )
            for i, out_size in enumerate(layer_sizes):
                layers.append(
                    SlimFC(
                        in_size=out_channels,
                        out_size=out_size,
                        activation_fn=post_fcnet_activation,
                        initializer=normc_initializer(1.0),
                    )
                )
                out_channels = out_size

        # Finish network normally (w/o overriding last layer size with
        # `num_outputs`), then add another linear one of size `num_outputs`.
        else:
            layers.append(
                SlimConv2d(
                    in_channels,
                    out_channels,
                    kernel,
                    stride,
                    None,  # padding=valid
                    activation_fn=activation,
                )
            )

            # num_outputs defined. Use that to create an exact
            # `num_output`-sized (1,1)-Conv2D.
            if num_outputs:
                in_size = [
                    np.ceil((in_size[0] - kernel[0]) / stride),
                    np.ceil((in_size[1] - kernel[1]) / stride),
                ]
                padding, _ = same_padding(in_size, [1, 1], [1, 1])
                if post_fcnet_hiddens:
                    layers.append(nn.Flatten())
                    in_size = out_channels
                    # Add (optional) post-fc-stack after last Conv2D layer.
                    for i, out_size in enumerate(post_fcnet_hiddens + [num_outputs]):
                        layers.append(
                            SlimFC(
                                in_size=in_size,
                                out_size=out_size,
                                activation_fn=post_fcnet_activation
                                if i < len(post_fcnet_hiddens) - 1
                                else None,
                                initializer=normc_initializer(1.0),
                            )
                        )
                        in_size = out_size
                    # Last layer is logits layer.
                    self._logits = layers.pop()

                else:
                    self._logits = SlimConv2d(
                        out_channels,
                        num_outputs,
                        [1, 1],
                        1,
                        padding,
                        activation_fn=None,
                    )

            # num_outputs not known -> Flatten, then set self.num_outputs
            # to the resulting number of nodes.
            else:
                self.last_layer_is_flattened = True
                layers.append(nn.Flatten())

        self._convs = nn.Sequential(*layers)

        # If our num_outputs still unknown, we need to do a test pass to
        # figure out the output dimensions. This could be the case, if we have
        # the Flatten layer at the end.
        if self.num_outputs is None:
            # Create a B=1 dummy sample and push it through out conv-net.
            dummy_in = (
                torch.from_numpy(self.obs_space.sample())
                .permute(2, 0, 1)
                .unsqueeze(0)
                .float()
            )
            dummy_out = self._convs(dummy_in)
            self.num_outputs = dummy_out.shape[1]

        # Build the value layers
        self._value_branch_separate = self._value_branch = None
        if vf_share_layers:
            self._value_branch = SlimFC(
                out_channels, 1, initializer=normc_initializer(0.01), activation_fn=None
            )
        else:
            vf_layers = []
            (w, h, in_channels) = obs_space.shape
            in_size = [w, h]
            for out_channels, kernel, stride in filters[:-1]:
                padding, out_size = same_padding(in_size, kernel, stride)
                vf_layers.append(
                    SlimConv2d(
                        in_channels,
                        out_channels,
                        kernel,
                        stride,
                        padding,
                        activation_fn=activation,
                    )
                )
                in_channels = out_channels
                in_size = out_size

            out_channels, kernel, stride = filters[-1]
            vf_layers.append(
                SlimConv2d(
                    in_channels,
                    out_channels,
                    kernel,
                    stride,
                    None,
                    activation_fn=activation,
                )
            )

            vf_layers.append(
                SlimConv2d(
                    in_channels=out_channels,
                    out_channels=1,
                    kernel=1,
                    stride=1,
                    padding=None,
                    activation_fn=None,
                )
            )
            self._value_branch_separate = nn.Sequential(*vf_layers)

        # Holds the current "base" output (before logits layer).
        self._features = None

    @override(TorchModelV2)
    def forward(
        self,
        input_dict: Dict[str, TensorType],
        state: List[TensorType],
        seq_lens: TensorType,
    ) -> (TensorType, List[TensorType]):
        self._features = input_dict["obs"].float()
        # Permuate b/c data comes in as [B, dim, dim, channels]:
        self._features = self._features.permute(0, 3, 1, 2)
        conv_out = self._convs(self._features)
        # Store features to save forward pass when getting value_function out.
        if not self._value_branch_separate:
            self._features = conv_out

        if not self.last_layer_is_flattened:
            if self._logits:
                conv_out = self._logits(conv_out)
            if len(conv_out.shape) == 4:
                if conv_out.shape[2] != 1 or conv_out.shape[3] != 1:
                    raise ValueError(
                        "Given `conv_filters` ({}) do not result in a [B, {} "
                        "(`num_outputs`), 1, 1] shape (but in {})! Please "
                        "adjust your Conv2D stack such that the last 2 dims "
                        "are both 1.".format(
                            self.model_config["conv_filters"],
                            self.num_outputs,
                            list(conv_out.shape),
                        )
                    )
                logits = conv_out.squeeze(3)
                logits = logits.squeeze(2)
            else:
                logits = conv_out
            return logits, state
        else:
            return conv_out, state

    @override(TorchModelV2)
    def value_function(self) -> TensorType:
        assert self._features is not None, "must call forward() first"
        if self._value_branch_separate:
            value = self._value_branch_separate(self._features)
            value = value.squeeze(3)
            value = value.squeeze(2)
            return value.squeeze(1)
        else:
            if not self.last_layer_is_flattened:
                features = self._features.squeeze(3)
                features = features.squeeze(2)
            else:
                features = self._features
            return self._value_branch(features).squeeze(1)

    def _hidden_layers(self, obs: TensorType) -> TensorType:
        res = self._convs(obs.permute(0, 3, 1, 2))  # switch to channel-major
        res = res.squeeze(3)
        res = res.squeeze(2)
        return res



import gymnasium as gym
import logging
import time
from typing import Dict

from gymnasium import spaces
import ray
from ray.rllib.algorithms.bandit.bandit_tf_model import (
    DiscreteLinearModelThompsonSampling,
    DiscreteLinearModelUCB,
    DiscreteLinearModel,
    ParametricLinearModelThompsonSampling,
    ParametricLinearModelUCB,
)
from ray.rllib.models.catalog import ModelCatalog
from ray.rllib.models.modelv2 import restore_original_dimensions
from ray.rllib.policy.policy import Policy
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.policy.tf_policy_template import build_tf_policy
from ray.rllib.utils.error import UnsupportedSpaceException
from ray.rllib.utils.metrics.learner_info import LEARNER_STATS_KEY
from ray.rllib.utils.tf_utils import make_tf_callable
from ray.rllib.utils.typing import TensorType, AlgorithmConfigDict
from ray.util.debug import log_once

logger = logging.getLogger(__name__)


class BanditPolicyOverrides:
    def __init__(self):
        @make_tf_callable(self.get_session())
        def learn_on_batch(postprocessed_batch) -> Dict[str, TensorType]:
            # INFOS dict can't be converted to Tensor with the interceptor.
            postprocessed_batch.set_get_interceptor(None)

            unflattened_obs = restore_original_dimensions(
                postprocessed_batch[SampleBatch.CUR_OBS],
                self.observation_space,
                self.framework,
            )

            info = {}

            start = time.time()
            self.model.partial_fit(
                unflattened_obs,
                postprocessed_batch[SampleBatch.REWARDS],
                postprocessed_batch[SampleBatch.ACTIONS],
            )

            infos = postprocessed_batch[SampleBatch.INFOS]
            if "regret" in infos[0]:
                regret = sum(
                    row["infos"]["regret"] for row in postprocessed_batch.rows()
                )
                self.regrets.append(regret)
                info["cumulative_regret"] = sum(self.regrets)
            else:
                if log_once("no_regrets"):
                    logger.warning(
                        "The env did not report `regret` values in "
                        "its `info` return, ignoring."
                    )
            info["update_latency"] = time.time() - start
            return {LEARNER_STATS_KEY: info}

        self.learn_on_batch = learn_on_batch


def validate_spaces(
    policy: Policy,
    observation_space: gym.spaces.Space,
    action_space: gym.spaces.Space,
    config: AlgorithmConfigDict,
) -> None:
    """Validates the observation- and action spaces used for the Policy.

    Args:
        policy: The policy, whose spaces are being validated.
        observation_space: The observation space to validate.
        action_space: The action space to validate.
        config: The Policy's config dict.

    Raises:
        UnsupportedSpaceException: If one of the spaces is not supported.
    """
    # Only support single Box or single Discrete spaces.
    if not isinstance(action_space, gym.spaces.Discrete):
        msg = (
            f"Action space ({action_space}) of {policy} is not supported for "
            f"Bandit algorithms. Must be `Discrete`."
        )
        # Hint at using the MultiDiscrete to Discrete wrapper for Bandits.
        if isinstance(action_space, gym.spaces.MultiDiscrete):
            msg += (
                " Try to wrap your environment with the "
                "`ray.rllib.env.wrappers.recsim::"
                "MultiDiscreteToDiscreteActionWrapper` class: `tune.register_env("
                "[some str], lambda ctx: MultiDiscreteToDiscreteActionWrapper("
                "[your gym env])); config = {'env': [some str]}`"
            )
        raise UnsupportedSpaceException(msg)


def make_model(policy, obs_space, action_space, config):
    _, logit_dim = ModelCatalog.get_action_dist(
        action_space, config["model"], framework="tf"
    )

    model_cls = DiscreteLinearModel

    if hasattr(obs_space, "original_space"):
        original_space = obs_space.original_space
    else:
        original_space = obs_space

    exploration_config = config.get("exploration_config")
    # Model is dependent on exploration strategy because of its implicitness

    # TODO: Have a separate model catalogue for bandits
    if exploration_config:
        if exploration_config["type"] == "ThompsonSampling":
            if isinstance(original_space, spaces.Dict):
                assert (
                    "item" in original_space.spaces
                ), "Cannot find 'item' key in observation space"
                model_cls = ParametricLinearModelThompsonSampling
            else:
                model_cls = DiscreteLinearModelThompsonSampling
        elif exploration_config["type"] == "UpperConfidenceBound":
            if isinstance(original_space, spaces.Dict):
                assert (
                    "item" in original_space.spaces
                ), "Cannot find 'item' key in observation space"
                model_cls = ParametricLinearModelUCB
            else:
                model_cls = DiscreteLinearModelUCB

    model = model_cls(
        obs_space, action_space, logit_dim, config["model"], name="LinearModel"
    )
    return model


def after_init(policy, *args):
    policy.regrets = []
    BanditPolicyOverrides.__init__(policy)


BanditTFPolicy = build_tf_policy(
    name="BanditTFPolicy",
    get_default_config=lambda: ray.rllib.algorithms.bandit.bandit.BanditConfig(),
    validate_spaces=validate_spaces,
    make_model=make_model,
    loss_fn=None,
    mixins=[BanditPolicyOverrides],
    after_init=after_init,
)



"""TensorFlow policy class used for R2D2."""

from typing import Dict, List, Optional, Tuple

import gymnasium as gym
import ray
from ray.rllib.algorithms.dqn.dqn_tf_policy import (
    clip_gradients,
    compute_q_values,
    PRIO_WEIGHTS,
    postprocess_nstep_and_prio,
)
from ray.rllib.algorithms.dqn.dqn_tf_policy import build_q_model
from ray.rllib.models.action_dist import ActionDistribution
from ray.rllib.models.modelv2 import ModelV2
from ray.rllib.models.tf.tf_action_dist import Categorical
from ray.rllib.models.torch.torch_action_dist import TorchCategorical
from ray.rllib.policy.policy import Policy
from ray.rllib.policy.tf_policy_template import build_tf_policy
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.policy.tf_mixins import LearningRateSchedule, TargetNetworkMixin
from ray.rllib.utils.framework import try_import_tf
from ray.rllib.utils.tf_utils import huber_loss
from ray.rllib.utils.typing import ModelInputDict, TensorType, AlgorithmConfigDict

tf1, tf, tfv = try_import_tf()


def build_r2d2_model(
    policy: Policy,
    obs_space: gym.spaces.Space,
    action_space: gym.spaces.Space,
    config: AlgorithmConfigDict,
) -> Tuple[ModelV2, ActionDistribution]:
    """Build q_model and target_model for DQN

    Args:
        policy: The policy, which will use the model for optimization.
        obs_space (gym.spaces.Space): The policy's observation space.
        action_space (gym.spaces.Space): The policy's action space.
        config (AlgorithmConfigDict):

    Returns:
        q_model
            Note: The target q model will not be returned, just assigned to
            `policy.target_model`.
    """

    # Create the policy's models.
    model = build_q_model(policy, obs_space, action_space, config)

    # Assert correct model type by checking the init state to be present.
    # For attention nets: These don't necessarily publish their init state via
    # Model.get_initial_state, but may only use the trajectory view API
    # (view_requirements).
    assert (
        model.get_initial_state() != []
        or model.view_requirements.get("state_in_0") is not None
    ), (
        "R2D2 requires its model to be a recurrent one! Try using "
        "`model.use_lstm` or `model.use_attention` in your config "
        "to auto-wrap your model with an LSTM- or attention net."
    )

    return model


def r2d2_loss(policy: Policy, model, _, train_batch: SampleBatch) -> TensorType:
    """Constructs the loss for R2D2TFPolicy.

    Args:
        policy: The Policy to calculate the loss for.
        model (ModelV2): The Model to calculate the loss for.
        train_batch: The training data.

    Returns:
        TensorType: A single loss tensor.
    """
    config = policy.config

    # Construct internal state inputs.
    i = 0
    state_batches = []
    while "state_in_{}".format(i) in train_batch:
        state_batches.append(train_batch["state_in_{}".format(i)])
        i += 1
    assert state_batches

    # Q-network evaluation (at t).
    q, _, _, _ = compute_q_values(
        policy,
        model,
        train_batch,
        state_batches=state_batches,
        seq_lens=train_batch.get(SampleBatch.SEQ_LENS),
        explore=False,
        is_training=True,
    )

    # Target Q-network evaluation (at t+1).
    q_target, _, _, _ = compute_q_values(
        policy,
        policy.target_model,
        train_batch,
        state_batches=state_batches,
        seq_lens=train_batch.get(SampleBatch.SEQ_LENS),
        explore=False,
        is_training=True,
    )

    if not hasattr(policy, "target_q_func_vars"):
        policy.target_q_func_vars = policy.target_model.variables()

    actions = tf.cast(train_batch[SampleBatch.ACTIONS], tf.int64)
    dones = tf.cast(train_batch[SampleBatch.TERMINATEDS], tf.float32)
    rewards = train_batch[SampleBatch.REWARDS]
    weights = tf.cast(train_batch[PRIO_WEIGHTS], tf.float32)

    B = tf.shape(state_batches[0])[0]
    T = tf.shape(q)[0] // B

    # Q scores for actions which we know were selected in the given state.
    one_hot_selection = tf.one_hot(actions, policy.action_space.n)
    q_selected = tf.reduce_sum(
        tf.where(q > tf.float32.min, q, tf.zeros_like(q)) * one_hot_selection, axis=1
    )

    if config["double_q"]:
        best_actions = tf.argmax(q, axis=1)
    else:
        best_actions = tf.argmax(q_target, axis=1)

    best_actions_one_hot = tf.one_hot(best_actions, policy.action_space.n)
    q_target_best = tf.reduce_sum(
        tf.where(q_target > tf.float32.min, q_target, tf.zeros_like(q_target))
        * best_actions_one_hot,
        axis=1,
    )

    if config["num_atoms"] > 1:
        raise ValueError("Distributional R2D2 not supported yet!")
    else:
        q_target_best_masked_tp1 = (1.0 - dones) * tf.concat(
            [q_target_best[1:], tf.constant([0.0])], axis=0
        )

        if config["use_h_function"]:
            h_inv = h_inverse(q_target_best_masked_tp1, config["h_function_epsilon"])
            target = h_function(
                rewards + config["gamma"] ** config["n_step"] * h_inv,
                config["h_function_epsilon"],
            )
        else:
            target = (
                rewards + config["gamma"] ** config["n_step"] * q_target_best_masked_tp1
            )

        # Seq-mask all loss-related terms.
        seq_mask = tf.sequence_mask(train_batch[SampleBatch.SEQ_LENS], T)[:, :-1]
        # Mask away also the burn-in sequence at the beginning.
        burn_in = policy.config["replay_buffer_config"]["replay_burn_in"]
        # Making sure, this works for both static graph and eager.
        if burn_in > 0:
            seq_mask = tf.cond(
                pred=tf.convert_to_tensor(burn_in, tf.int32) < T,
                true_fn=lambda: tf.concat(
                    [tf.fill([B, burn_in], False), seq_mask[:, burn_in:]], 1
                ),
                false_fn=lambda: seq_mask,
            )

        def reduce_mean_valid(t):
            return tf.reduce_mean(tf.boolean_mask(t, seq_mask))

        # Make sure to use the correct time indices:
        # Q(t) - [gamma * r + Q^(t+1)]
        q_selected = tf.reshape(q_selected, [B, T])[:, :-1]
        td_error = q_selected - tf.stop_gradient(tf.reshape(target, [B, T])[:, :-1])
        td_error = td_error * tf.cast(seq_mask, tf.float32)
        weights = tf.reshape(weights, [B, T])[:, :-1]
        policy._total_loss = reduce_mean_valid(weights * huber_loss(td_error))
        # Store the TD-error per time chunk (b/c we need only one mean
        # prioritized replay weight per stored sequence).
        policy._td_error = tf.reduce_mean(td_error, axis=-1)
        policy._loss_stats = {
            "mean_q": reduce_mean_valid(q_selected),
            "min_q": tf.reduce_min(q_selected),
            "max_q": tf.reduce_max(q_selected),
            "mean_td_error": reduce_mean_valid(td_error),
        }

    return policy._total_loss


def h_function(x, epsilon=1.0):
    """h-function to normalize target Qs, described in the paper [1].

    h(x) = sign(x) * [sqrt(abs(x) + 1) - 1] + epsilon * x

    Used in [1] in combination with h_inverse:
      targets = h(r + gamma * h_inverse(Q^))
    """
    return tf.sign(x) * (tf.sqrt(tf.abs(x) + 1.0) - 1.0) + epsilon * x


def h_inverse(x, epsilon=1.0):
    """Inverse if the above h-function, described in the paper [1].

    If x > 0.0:
    h-1(x) = [2eps * x + (2eps + 1) - sqrt(4eps x + (2eps + 1)^2)] /
        (2 * eps^2)

    If x < 0.0:
    h-1(x) = [2eps * x + (2eps + 1) + sqrt(-4eps x + (2eps + 1)^2)] /
        (2 * eps^2)
    """
    two_epsilon = epsilon * 2
    if_x_pos = (
        two_epsilon * x
        + (two_epsilon + 1.0)
        - tf.sqrt(4.0 * epsilon * x + (two_epsilon + 1.0) ** 2)
    ) / (2.0 * epsilon**2)
    if_x_neg = (
        two_epsilon * x
        - (two_epsilon + 1.0)
        + tf.sqrt(-4.0 * epsilon * x + (two_epsilon + 1.0) ** 2)
    ) / (2.0 * epsilon**2)
    return tf.where(x < 0.0, if_x_neg, if_x_pos)


class ComputeTDErrorMixin:
    """Assign the `compute_td_error` method to the R2D2TFPolicy

    This allows us to prioritize on the worker side.
    """

    def __init__(self):
        def compute_td_error(
            obs_t, act_t, rew_t, obs_tp1, terminateds_mask, importance_weights
        ):
            input_dict = self._lazy_tensor_dict({SampleBatch.CUR_OBS: obs_t})
            input_dict[SampleBatch.ACTIONS] = act_t
            input_dict[SampleBatch.REWARDS] = rew_t
            input_dict[SampleBatch.NEXT_OBS] = obs_tp1
            input_dict[SampleBatch.TERMINATEDS] = terminateds_mask
            input_dict[PRIO_WEIGHTS] = importance_weights

            # Do forward pass on loss to update td error attribute
            r2d2_loss(self, self.model, None, input_dict)

            return self._td_error

        self.compute_td_error = compute_td_error


def get_distribution_inputs_and_class(
    policy: Policy,
    model: ModelV2,
    *,
    input_dict: ModelInputDict,
    state_batches: Optional[List[TensorType]] = None,
    seq_lens: Optional[TensorType] = None,
    explore: bool = True,
    is_training: bool = False,
    **kwargs
) -> Tuple[TensorType, type, List[TensorType]]:

    if policy.config["framework"] == "torch":
        from ray.rllib.algorithms.r2d2.r2d2_torch_policy import (
            compute_q_values as torch_compute_q_values,
        )

        func = torch_compute_q_values
    else:
        func = compute_q_values

    q_vals, logits, probs_or_logits, state_out = func(
        policy, model, input_dict, state_batches, seq_lens, explore, is_training
    )

    policy.q_values = q_vals
    if not hasattr(policy, "q_func_vars"):
        policy.q_func_vars = model.variables()

    action_dist_class = (
        TorchCategorical if policy.config["framework"] == "torch" else Categorical
    )

    return policy.q_values, action_dist_class, state_out


def adam_optimizer(
    policy: Policy, config: AlgorithmConfigDict
) -> "tf.keras.optimizers.Optimizer":
    return tf1.train.AdamOptimizer(
        learning_rate=policy.cur_lr, epsilon=config["adam_epsilon"]
    )


def build_q_stats(policy: Policy, batch) -> Dict[str, TensorType]:
    return dict(
        {
            "cur_lr": policy.cur_lr,
        },
        **policy._loss_stats
    )


def setup_early_mixins(
    policy: Policy, obs_space, action_space, config: AlgorithmConfigDict
) -> None:
    LearningRateSchedule.__init__(policy, config["lr"], config["lr_schedule"])


def before_loss_init(
    policy: Policy,
    obs_space: gym.spaces.Space,
    action_space: gym.spaces.Space,
    config: AlgorithmConfigDict,
) -> None:
    ComputeTDErrorMixin.__init__(policy)


def setup_late_mixins(
    policy: Policy,
    obs_space: gym.spaces.Space,
    action_space: gym.spaces.Space,
    config: AlgorithmConfigDict,
) -> None:
    TargetNetworkMixin.__init__(policy)


R2D2TFPolicy = build_tf_policy(
    name="R2D2TFPolicy",
    loss_fn=r2d2_loss,
    get_default_config=lambda: ray.rllib.algorithms.r2d2.r2d2.R2D2Config(),
    postprocess_fn=postprocess_nstep_and_prio,
    stats_fn=build_q_stats,
    make_model=build_r2d2_model,
    action_distribution_fn=get_distribution_inputs_and_class,
    optimizer_fn=adam_optimizer,
    extra_action_out_fn=lambda policy: {"q_values": policy.q_values},
    compute_gradients_fn=clip_gradients,
    extra_learn_fetches_fn=lambda policy: {"td_error": policy._td_error},
    before_init=setup_early_mixins,
    before_loss_init=before_loss_init,
    after_init=setup_late_mixins,
    mixins=[
        TargetNetworkMixin,
        ComputeTDErrorMixin,
        LearningRateSchedule,
    ],
)



from typing import Optional

import tree

from ray.rllib.core.models.base import (
    Encoder,
    ActorCriticEncoder,
    StatefulActorCriticEncoder,
    STATE_IN,
    STATE_OUT,
    ENCODER_OUT,
)
from ray.rllib.core.models.base import Model, tokenize
from ray.rllib.core.models.configs import (
    ActorCriticEncoderConfig,
    CNNEncoderConfig,
    MLPEncoderConfig,
    RecurrentEncoderConfig,
)
from ray.rllib.core.models.specs.specs_base import Spec
from ray.rllib.core.models.specs.specs_base import TensorSpec
from ray.rllib.core.models.specs.specs_dict import SpecDict
from ray.rllib.core.models.torch.base import TorchModel
from ray.rllib.core.models.torch.primitives import TorchMLP, TorchCNN
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.utils.annotations import override
from ray.rllib.utils.framework import try_import_torch

torch, nn = try_import_torch()


class TorchActorCriticEncoder(TorchModel, ActorCriticEncoder):
    """An actor-critic encoder for torch."""

    framework = "torch"

    def __init__(self, config: ActorCriticEncoderConfig) -> None:
        TorchModel.__init__(self, config)
        ActorCriticEncoder.__init__(self, config)


class TorchStatefulActorCriticEncoder(TorchModel, StatefulActorCriticEncoder):
    """A stateful actor-critic encoder for torch."""

    framework = "torch"

    def __init__(self, config: ActorCriticEncoderConfig) -> None:
        TorchModel.__init__(self, config)
        StatefulActorCriticEncoder.__init__(self, config)


class TorchMLPEncoder(TorchModel, Encoder):
    def __init__(self, config: MLPEncoderConfig) -> None:
        TorchModel.__init__(self, config)
        Encoder.__init__(self, config)

        # Create the neural network.
        self.net = TorchMLP(
            input_dim=config.input_dims[0],
            hidden_layer_dims=config.hidden_layer_dims,
            hidden_layer_activation=config.hidden_layer_activation,
            hidden_layer_use_layernorm=config.hidden_layer_use_layernorm,
            hidden_layer_use_bias=config.hidden_layer_use_bias,
            output_dim=config.output_layer_dim,
            output_activation=config.output_layer_activation,
            output_use_bias=config.output_layer_use_bias,
        )

    @override(Model)
    def get_input_specs(self) -> Optional[Spec]:
        return SpecDict(
            {
                SampleBatch.OBS: TensorSpec(
                    "b, d", d=self.config.input_dims[0], framework="torch"
                ),
            }
        )

    @override(Model)
    def get_output_specs(self) -> Optional[Spec]:
        return SpecDict(
            {
                ENCODER_OUT: TensorSpec(
                    "b, d", d=self.config.output_dims[0], framework="torch"
                ),
            }
        )

    @override(Model)
    def _forward(self, inputs: dict, **kwargs) -> dict:
        return {ENCODER_OUT: self.net(inputs[SampleBatch.OBS])}


class TorchCNNEncoder(TorchModel, Encoder):
    def __init__(self, config: CNNEncoderConfig) -> None:
        TorchModel.__init__(self, config)
        Encoder.__init__(self, config)

        layers = []
        # The bare-bones CNN (no flatten, no succeeding dense).
        cnn = TorchCNN(
            input_dims=config.input_dims,
            cnn_filter_specifiers=config.cnn_filter_specifiers,
            cnn_activation=config.cnn_activation,
            cnn_use_layernorm=config.cnn_use_layernorm,
            cnn_use_bias=config.cnn_use_bias,
        )
        layers.append(cnn)

        # Add a flatten operation to move from 2/3D into 1D space.
        if config.flatten_at_end:
            layers.append(nn.Flatten())

        # Create the network from gathered layers.
        self.net = nn.Sequential(*layers)

    @override(Model)
    def get_input_specs(self) -> Optional[Spec]:
        return SpecDict(
            {
                SampleBatch.OBS: TensorSpec(
                    "b, w, h, c",
                    w=self.config.input_dims[0],
                    h=self.config.input_dims[1],
                    c=self.config.input_dims[2],
                    framework="torch",
                ),
            }
        )

    @override(Model)
    def get_output_specs(self) -> Optional[Spec]:
        return SpecDict(
            {
                ENCODER_OUT: (
                    TensorSpec("b, d", d=self.config.output_dims[0], framework="torch")
                    if self.config.flatten_at_end
                    else TensorSpec(
                        "b, w, h, c",
                        w=self.config.output_dims[0],
                        h=self.config.output_dims[1],
                        d=self.config.output_dims[2],
                        framework="torch",
                    )
                )
            }
        )

    @override(Model)
    def _forward(self, inputs: dict, **kwargs) -> dict:
        return {ENCODER_OUT: self.net(inputs[SampleBatch.OBS])}


class TorchGRUEncoder(TorchModel, Encoder):
    """A recurrent GRU encoder.

    This encoder has...
    - Zero or one tokenizers.
    - One or more GRU layers.
    - One linear output layer.
    """

    def __init__(self, config: RecurrentEncoderConfig) -> None:
        TorchModel.__init__(self, config)

        # Maybe create a tokenizer
        if config.tokenizer_config is not None:
            self.tokenizer = config.tokenizer_config.build(framework="torch")
            gru_input_dims = config.tokenizer_config.output_dims
        else:
            self.tokenizer = None
            gru_input_dims = config.input_dims

        # We only support 1D spaces right now.
        assert len(gru_input_dims) == 1
        gru_input_dim = gru_input_dims[0]

        # Create the torch LSTM layer.
        self.gru = nn.GRU(
            gru_input_dim,
            config.hidden_dim,
            config.num_layers,
            batch_first=config.batch_major,
            bias=config.use_bias,
        )

    @override(Model)
    def get_input_specs(self) -> Optional[Spec]:
        return SpecDict(
            {
                # b, t for batch major; t, b for time major.
                SampleBatch.OBS: TensorSpec(
                    "b, t, d",
                    d=self.config.input_dims[0],
                    framework="torch",
                ),
                STATE_IN: {
                    "h": TensorSpec(
                        "b, l, h",
                        h=self.config.hidden_dim,
                        l=self.config.num_layers,
                        framework="torch",
                    ),
                },
            }
        )

    @override(Model)
    def get_output_specs(self) -> Optional[Spec]:
        return SpecDict(
            {
                ENCODER_OUT: TensorSpec(
                    "b, t, d", d=self.config.output_dims[0], framework="torch"
                ),
                STATE_OUT: {
                    "h": TensorSpec(
                        "b, l, h",
                        h=self.config.hidden_dim,
                        l=self.config.num_layers,
                        framework="torch",
                    ),
                },
            }
        )

    @override(Model)
    def get_initial_state(self):
        return {
            "h": torch.zeros(self.config.num_layers, self.config.hidden_dim),
        }

    @override(Model)
    def _forward(self, inputs: dict, **kwargs) -> dict:
        outputs = {}

        if self.tokenizer is not None:
            # Push observations through the tokenizer encoder if we built one.
            out = tokenize(self.tokenizer, inputs, framework="torch")
        else:
            # Otherwise, just use the raw observations.
            out = inputs[SampleBatch.OBS].float()

        # States are batch-first when coming in. Make them layers-first.
        states_in = tree.map_structure(lambda s: s.transpose(0, 1), inputs[STATE_IN])

        out, states_out = self.gru(out, states_in["h"])
        states_out = {"h": states_out}

        # Insert them into the output dict.
        outputs[ENCODER_OUT] = out
        outputs[STATE_OUT] = tree.map_structure(lambda s: s.transpose(0, 1), states_out)
        return outputs


class TorchLSTMEncoder(TorchModel, Encoder):
    """A recurrent LSTM encoder.

    This encoder has...
    - Zero or one tokenizers.
    - One or more LSTM layers.
    - One linear output layer.
    """

    def __init__(self, config: RecurrentEncoderConfig) -> None:
        TorchModel.__init__(self, config)

        # Maybe create a tokenizer
        if config.tokenizer_config is not None:
            self.tokenizer = config.tokenizer_config.build(framework="torch")
            lstm_input_dims = config.tokenizer_config.output_dims
        else:
            self.tokenizer = None
            lstm_input_dims = config.input_dims

        # We only support 1D spaces right now.
        assert len(lstm_input_dims) == 1
        lstm_input_dim = lstm_input_dims[0]

        # Create the torch LSTM layer.
        self.lstm = nn.LSTM(
            lstm_input_dim,
            config.hidden_dim,
            config.num_layers,
            batch_first=config.batch_major,
            bias=config.use_bias,
        )

    @override(Model)
    def get_input_specs(self) -> Optional[Spec]:
        return SpecDict(
            {
                # b, t for batch major; t, b for time major.
                SampleBatch.OBS: TensorSpec(
                    "b, t, d", d=self.config.input_dims[0], framework="torch"
                ),
                STATE_IN: {
                    "h": TensorSpec(
                        "b, l, h",
                        h=self.config.hidden_dim,
                        l=self.config.num_layers,
                        framework="torch",
                    ),
                    "c": TensorSpec(
                        "b, l, h",
                        h=self.config.hidden_dim,
                        l=self.config.num_layers,
                        framework="torch",
                    ),
                },
            }
        )

    @override(Model)
    def get_output_specs(self) -> Optional[Spec]:
        return SpecDict(
            {
                ENCODER_OUT: TensorSpec(
                    "b, t, d", d=self.config.output_dims[0], framework="torch"
                ),
                STATE_OUT: {
                    "h": TensorSpec(
                        "b, l, h",
                        h=self.config.hidden_dim,
                        l=self.config.num_layers,
                        framework="torch",
                    ),
                    "c": TensorSpec(
                        "b, l, h",
                        h=self.config.hidden_dim,
                        l=self.config.num_layers,
                        framework="torch",
                    ),
                },
            }
        )

    @override(Model)
    def get_initial_state(self):
        return {
            "h": torch.zeros(self.config.num_layers, self.config.hidden_dim),
            "c": torch.zeros(self.config.num_layers, self.config.hidden_dim),
        }

    @override(Model)
    def _forward(self, inputs: dict, **kwargs) -> dict:
        outputs = {}

        if self.tokenizer is not None:
            # Push observations through the tokenizer encoder if we built one.
            out = tokenize(self.tokenizer, inputs, framework="torch")
        else:
            # Otherwise, just use the raw observations.
            out = inputs[SampleBatch.OBS].float()

        # States are batch-first when coming in. Make them layers-first.
        states_in = tree.map_structure(lambda s: s.transpose(0, 1), inputs[STATE_IN])

        out, states_out = self.lstm(out, (states_in["h"], states_in["c"]))
        states_out = {"h": states_out[0], "c": states_out[1]}

        # Insert them into the output dict.
        outputs[ENCODER_OUT] = out
        outputs[STATE_OUT] = tree.map_structure(lambda s: s.transpose(0, 1), states_out)
        return outputs



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/ray_client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!src/ray/protobuf/ray_client.proto\x12\x07ray.rpc\"\xb5\x01\n\x03\x41rg\x12+\n\x05local\x18\x01 \x01(\x0e\x32\x15.ray.rpc.Arg.LocalityR\x05local\x12!\n\x0creference_id\x18\x02 \x01(\x0cR\x0breferenceId\x12\x12\n\x04\x64\x61ta\x18\x03 \x01(\x0cR\x04\x64\x61ta\x12!\n\x04type\x18\x04 \x01(\x0e\x32\r.ray.rpc.TypeR\x04type\"\'\n\x08Locality\x12\x0c\n\x08INTERNED\x10\x00\x12\r\n\tREFERENCE\x10\x01\"6\n\x0bTaskOptions\x12\'\n\x0fpickled_options\x18\x01 \x01(\x0cR\x0epickledOptions\"\xf4\x04\n\nClientTask\x12\x36\n\x04type\x18\x01 \x01(\x0e\x32\".ray.rpc.ClientTask.RemoteExecTypeR\x04type\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x1d\n\npayload_id\x18\x03 \x01(\x0cR\tpayloadId\x12 \n\x04\x61rgs\x18\x04 \x03(\x0b\x32\x0c.ray.rpc.ArgR\x04\x61rgs\x12\x37\n\x06kwargs\x18\x05 \x03(\x0b\x32\x1f.ray.rpc.ClientTask.KwargsEntryR\x06kwargs\x12\x1b\n\tclient_id\x18\x06 \x01(\tR\x08\x63lientId\x12.\n\x07options\x18\x07 \x01(\x0b\x32\x14.ray.rpc.TaskOptionsR\x07options\x12?\n\x10\x62\x61seline_options\x18\x08 \x01(\x0b\x32\x14.ray.rpc.TaskOptionsR\x0f\x62\x61selineOptions\x12\x1c\n\tnamespace\x18\t \x01(\tR\tnamespace\x12\x12\n\x04\x64\x61ta\x18\n \x01(\x0cR\x04\x64\x61ta\x12\x19\n\x08\x63hunk_id\x18\x0b \x01(\x05R\x07\x63hunkId\x12!\n\x0ctotal_chunks\x18\x0c \x01(\x05R\x0btotalChunks\x1aG\n\x0bKwargsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\"\n\x05value\x18\x02 \x01(\x0b\x32\x0c.ray.rpc.ArgR\x05value:\x02\x38\x01\"Y\n\x0eRemoteExecType\x12\x0c\n\x08\x46UNCTION\x10\x00\x12\t\n\x05\x41\x43TOR\x10\x01\x12\n\n\x06METHOD\x10\x02\x12\x11\n\rSTATIC_METHOD\x10\x03\x12\x0f\n\x0bNAMED_ACTOR\x10\x04\"]\n\x10\x43lientTaskTicket\x12\x14\n\x05valid\x18\x01 \x01(\x08R\x05valid\x12\x1d\n\nreturn_ids\x18\x02 \x03(\x0cR\treturnIds\x12\x14\n\x05\x65rror\x18\x03 \x01(\x0cR\x05\x65rror\"\xbc\x01\n\nPutRequest\x12\x12\n\x04\x64\x61ta\x18\x01 \x01(\x0cR\x04\x64\x61ta\x12\"\n\rclient_ref_id\x18\x02 \x01(\x0cR\x0b\x63lientRefId\x12\x19\n\x08\x63hunk_id\x18\x03 \x01(\x05R\x07\x63hunkId\x12!\n\x0ctotal_chunks\x18\x04 \x01(\x05R\x0btotalChunks\x12\x1d\n\ntotal_size\x18\x05 \x01(\x03R\ttotalSize\x12\x19\n\x08owner_id\x18\x06 \x01(\x0cR\x07ownerId\"I\n\x0bPutResponse\x12\x0e\n\x02id\x18\x01 \x01(\x0cR\x02id\x12\x14\n\x05valid\x18\x02 \x01(\x08R\x05valid\x12\x14\n\x05\x65rror\x18\x03 \x01(\x0cR\x05\x65rror\"\x96\x01\n\nGetRequest\x12\x10\n\x03ids\x18\x04 \x03(\x0cR\x03ids\x12\x18\n\x07timeout\x18\x02 \x01(\x02R\x07timeout\x12\"\n\x0c\x61synchronous\x18\x03 \x01(\x08R\x0c\x61synchronous\x12$\n\x0estart_chunk_id\x18\x05 \x01(\x05R\x0cstartChunkId\x12\x12\n\x02id\x18\x01 \x01(\x0c\x42\x02\x18\x01R\x02id\"\xaa\x01\n\x0bGetResponse\x12\x14\n\x05valid\x18\x01 \x01(\x08R\x05valid\x12\x12\n\x04\x64\x61ta\x18\x02 \x01(\x0cR\x04\x64\x61ta\x12\x14\n\x05\x65rror\x18\x03 \x01(\x0cR\x05\x65rror\x12\x19\n\x08\x63hunk_id\x18\x04 \x01(\x05R\x07\x63hunkId\x12!\n\x0ctotal_chunks\x18\x05 \x01(\x05R\x0btotalChunks\x12\x1d\n\ntotal_size\x18\x06 \x01(\x04R\ttotalSize\"\x84\x01\n\x0bWaitRequest\x12\x1d\n\nobject_ids\x18\x01 \x03(\x0cR\tobjectIds\x12\x1f\n\x0bnum_returns\x18\x02 \x01(\x03R\nnumReturns\x12\x18\n\x07timeout\x18\x03 \x01(\x01R\x07timeout\x12\x1b\n\tclient_id\x18\x04 \x01(\tR\x08\x63lientId\"\x80\x01\n\x0cWaitResponse\x12\x14\n\x05valid\x18\x01 \x01(\x08R\x05valid\x12(\n\x10ready_object_ids\x18\x02 \x03(\x0cR\x0ereadyObjectIds\x12\x30\n\x14remaining_object_ids\x18\x03 \x03(\x0cR\x12remainingObjectIds\"\xad\x01\n\x0f\x43lusterInfoType\"\x99\x01\n\x08TypeEnum\x12\x12\n\x0eIS_INITIALIZED\x10\x00\x12\t\n\x05NODES\x10\x01\x12\x15\n\x11\x43LUSTER_RESOURCES\x10\x02\x12\x17\n\x13\x41VAILABLE_RESOURCES\x10\x03\x12\x13\n\x0fRUNTIME_CONTEXT\x10\x04\x12\x0c\n\x08TIMELINE\x10\x05\x12\x08\n\x04PING\x10\x06\x12\x11\n\rDASHBOARD_URL\x10\x07\"K\n\x12\x43lusterInfoRequest\x12\x35\n\x04type\x18\x01 \x01(\x0e\x32!.ray.rpc.ClusterInfoType.TypeEnumR\x04type\"\x8e\x05\n\x13\x43lusterInfoResponse\x12\x35\n\x04type\x18\x01 \x01(\x0e\x32!.ray.rpc.ClusterInfoType.TypeEnumR\x04type\x12\x14\n\x04json\x18\x02 \x01(\tH\x00R\x04json\x12S\n\x0eresource_table\x18\x03 \x01(\x0b\x32*.ray.rpc.ClusterInfoResponse.ResourceTableH\x00R\rresourceTable\x12V\n\x0fruntime_context\x18\x04 \x01(\x0b\x32+.ray.rpc.ClusterInfoResponse.RuntimeContextH\x00R\x0eruntimeContext\x1a\x96\x01\n\rResourceTable\x12K\n\x05table\x18\x01 \x03(\x0b\x32\x35.ray.rpc.ClusterInfoResponse.ResourceTable.TableEntryR\x05table\x1a\x38\n\nTableEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1a\xd2\x01\n\x0eRuntimeContext\x12\x15\n\x06job_id\x18\x01 \x01(\x0cR\x05jobId\x12\x17\n\x07node_id\x18\x02 \x01(\x0cR\x06nodeId\x12\x1c\n\tnamespace\x18\x03 \x01(\tR\tnamespace\x12\x1f\n\x0bruntime_env\x18\x04 \x01(\tR\nruntimeEnv\x12\x30\n\x14\x63\x61pture_client_tasks\x18\x05 \x01(\x08R\x12\x63\x61ptureClientTasks\x12\x1f\n\x0bgcs_address\x18\x06 \x01(\tR\ngcsAddressB\x0f\n\rresponse_type\"\xf1\x02\n\x10TerminateRequest\x12\x1b\n\tclient_id\x18\x01 \x01(\tR\x08\x63lientId\x12@\n\x05\x61\x63tor\x18\x02 \x01(\x0b\x32(.ray.rpc.TerminateRequest.ActorTerminateH\x00R\x05\x61\x63tor\x12P\n\x0btask_object\x18\x03 \x01(\x0b\x32-.ray.rpc.TerminateRequest.TaskObjectTerminateH\x00R\ntaskObject\x1a?\n\x0e\x41\x63torTerminate\x12\x0e\n\x02id\x18\x01 \x01(\x0cR\x02id\x12\x1d\n\nno_restart\x18\x02 \x01(\x08R\tnoRestart\x1aY\n\x13TaskObjectTerminate\x12\x0e\n\x02id\x18\x01 \x01(\x0cR\x02id\x12\x14\n\x05\x66orce\x18\x02 \x01(\x08R\x05\x66orce\x12\x1c\n\trecursive\x18\x03 \x01(\x08R\trecursiveB\x10\n\x0eterminate_type\"#\n\x11TerminateResponse\x12\x0e\n\x02ok\x18\x01 \x01(\x08R\x02ok\"T\n\x0fKVExistsRequest\x12\x10\n\x03key\x18\x01 \x01(\x0cR\x03key\x12!\n\tnamespace\x18\x02 \x01(\x0cH\x00R\tnamespace\x88\x01\x01\x42\x0c\n\n_namespace\"*\n\x10KVExistsResponse\x12\x16\n\x06\x65xists\x18\x01 \x01(\x08R\x06\x65xists\"Q\n\x0cKVGetRequest\x12\x10\n\x03key\x18\x01 \x01(\x0cR\x03key\x12!\n\tnamespace\x18\x02 \x01(\x0cH\x00R\tnamespace\x88\x01\x01\x42\x0c\n\n_namespace\"4\n\rKVGetResponse\x12\x19\n\x05value\x18\x01 \x01(\x0cH\x00R\x05value\x88\x01\x01\x42\x08\n\x06_value\"\x85\x01\n\x0cKVPutRequest\x12\x10\n\x03key\x18\x01 \x01(\x0cR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value\x12\x1c\n\toverwrite\x18\x03 \x01(\x08R\toverwrite\x12!\n\tnamespace\x18\x04 \x01(\x0cH\x00R\tnamespace\x88\x01\x01\x42\x0c\n\n_namespace\"6\n\rKVPutResponse\x12%\n\x0e\x61lready_exists\x18\x01 \x01(\x08R\ralreadyExists\"u\n\x0cKVDelRequest\x12\x10\n\x03key\x18\x01 \x01(\x0cR\x03key\x12\"\n\rdel_by_prefix\x18\x02 \x01(\x08R\x0b\x64\x65lByPrefix\x12!\n\tnamespace\x18\x03 \x01(\x0cH\x00R\tnamespace\x88\x01\x01\x42\x0c\n\n_namespace\"0\n\rKVDelResponse\x12\x1f\n\x0b\x64\x65leted_num\x18\x01 \x01(\x05R\ndeletedNum\"X\n\rKVListRequest\x12\x16\n\x06prefix\x18\x01 \x01(\x0cR\x06prefix\x12!\n\tnamespace\x18\x02 \x01(\x0cH\x00R\tnamespace\x88\x01\x01\x42\x0c\n\n_namespace\"$\n\x0eKVListResponse\x12\x12\n\x04keys\x18\x01 \x03(\x0cR\x04keys\"T\n\x1d\x43lientPinRuntimeEnvURIRequest\x12\x10\n\x03uri\x18\x01 \x01(\tR\x03uri\x12!\n\x0c\x65xpiration_s\x18\x02 \x01(\x05R\x0b\x65xpirationS\" \n\x1e\x43lientPinRuntimeEnvURIResponse\"\x8a\x01\n\x0bInitRequest\x12\x1d\n\njob_config\x18\x01 \x01(\x0cR\tjobConfig\x12&\n\x0fray_init_kwargs\x18\x02 \x01(\tR\rrayInitKwargs\x12\x34\n\x16reconnect_grace_period\x18\x03 \x01(\x05R\x14reconnectGracePeriod\"0\n\x0cInitResponse\x12\x0e\n\x02ok\x18\x01 \x01(\x08R\x02ok\x12\x10\n\x03msg\x18\x02 \x01(\tR\x03msg\"\x17\n\x15PrepRuntimeEnvRequest\"\x18\n\x16PrepRuntimeEnvResponse\"E\n\x1c\x43lientListNamedActorsRequest\x12%\n\x0e\x61ll_namespaces\x18\x01 \x01(\x08R\rallNamespaces\"@\n\x1d\x43lientListNamedActorsResponse\x12\x1f\n\x0b\x61\x63tors_json\x18\x01 \x01(\tR\nactorsJson\"\"\n\x0eReleaseRequest\x12\x10\n\x03ids\x18\x01 \x03(\x0cR\x03ids\"!\n\x0fReleaseResponse\x12\x0e\n\x02ok\x18\x02 \x03(\x08R\x02ok\"\x17\n\x15\x43onnectionInfoRequest\"\xcb\x01\n\x16\x43onnectionInfoResponse\x12\x1f\n\x0bnum_clients\x18\x01 \x01(\x05R\nnumClients\x12\x1f\n\x0bray_version\x18\x02 \x01(\tR\nrayVersion\x12\x1d\n\nray_commit\x18\x03 \x01(\tR\trayCommit\x12%\n\x0epython_version\x18\x04 \x01(\tR\rpythonVersion\x12)\n\x10protocol_version\x18\x05 \x01(\tR\x0fprotocolVersion\"\x1a\n\x18\x43onnectionCleanupRequest\"\x1b\n\x19\x43onnectionCleanupResponse\"+\n\x12\x41\x63knowledgeRequest\x12\x15\n\x06req_id\x18\x01 \x01(\x05R\x05reqId\"\xc6\x05\n\x0b\x44\x61taRequest\x12\x15\n\x06req_id\x18\x01 \x01(\x05R\x05reqId\x12\'\n\x03get\x18\x02 \x01(\x0b\x32\x13.ray.rpc.GetRequestH\x00R\x03get\x12\'\n\x03put\x18\x03 \x01(\x0b\x32\x13.ray.rpc.PutRequestH\x00R\x03put\x12\x33\n\x07release\x18\x04 \x01(\x0b\x32\x17.ray.rpc.ReleaseRequestH\x00R\x07release\x12I\n\x0f\x63onnection_info\x18\x05 \x01(\x0b\x32\x1e.ray.rpc.ConnectionInfoRequestH\x00R\x0e\x63onnectionInfo\x12*\n\x04init\x18\x06 \x01(\x0b\x32\x14.ray.rpc.InitRequestH\x00R\x04init\x12J\n\x10prep_runtime_env\x18\x07 \x01(\x0b\x32\x1e.ray.rpc.PrepRuntimeEnvRequestH\x00R\x0eprepRuntimeEnv\x12R\n\x12\x63onnection_cleanup\x18\x08 \x01(\x0b\x32!.ray.rpc.ConnectionCleanupRequestH\x00R\x11\x63onnectionCleanup\x12?\n\x0b\x61\x63knowledge\x18\t \x01(\x0b\x32\x1b.ray.rpc.AcknowledgeRequestH\x00R\x0b\x61\x63knowledge\x12)\n\x04task\x18\n \x01(\x0b\x32\x13.ray.rpc.ClientTaskH\x00R\x04task\x12\x39\n\tterminate\x18\x0b \x01(\x0b\x32\x19.ray.rpc.TerminateRequestH\x00R\tterminate\x12S\n\x11list_named_actors\x18\x0c \x01(\x0b\x32%.ray.rpc.ClientListNamedActorsRequestH\x00R\x0flistNamedActorsB\x06\n\x04type\"\xb5\x05\n\x0c\x44\x61taResponse\x12\x15\n\x06req_id\x18\x01 \x01(\x05R\x05reqId\x12(\n\x03get\x18\x02 \x01(\x0b\x32\x14.ray.rpc.GetResponseH\x00R\x03get\x12(\n\x03put\x18\x03 \x01(\x0b\x32\x14.ray.rpc.PutResponseH\x00R\x03put\x12\x34\n\x07release\x18\x04 \x01(\x0b\x32\x18.ray.rpc.ReleaseResponseH\x00R\x07release\x12J\n\x0f\x63onnection_info\x18\x05 \x01(\x0b\x32\x1f.ray.rpc.ConnectionInfoResponseH\x00R\x0e\x63onnectionInfo\x12+\n\x04init\x18\x06 \x01(\x0b\x32\x15.ray.rpc.InitResponseH\x00R\x04init\x12K\n\x10prep_runtime_env\x18\x07 \x01(\x0b\x32\x1f.ray.rpc.PrepRuntimeEnvResponseH\x00R\x0eprepRuntimeEnv\x12S\n\x12\x63onnection_cleanup\x18\x08 \x01(\x0b\x32\".ray.rpc.ConnectionCleanupResponseH\x00R\x11\x63onnectionCleanup\x12<\n\x0btask_ticket\x18\n \x01(\x0b\x32\x19.ray.rpc.ClientTaskTicketH\x00R\ntaskTicket\x12:\n\tterminate\x18\x0b \x01(\x0b\x32\x1a.ray.rpc.TerminateResponseH\x00R\tterminate\x12T\n\x11list_named_actors\x18\x0c \x01(\x0b\x32&.ray.rpc.ClientListNamedActorsResponseH\x00R\x0flistNamedActorsB\x06\n\x04typeJ\x04\x08\t\x10\nR\x0b\x61\x63knowledge\"J\n\x12LogSettingsRequest\x12\x18\n\x07\x65nabled\x18\x01 \x01(\x08R\x07\x65nabled\x12\x1a\n\x08loglevel\x18\x02 \x01(\x05R\x08loglevel\"E\n\x07LogData\x12\x10\n\x03msg\x18\x01 \x01(\tR\x03msg\x12\x14\n\x05level\x18\x02 \x01(\x05R\x05level\x12\x12\n\x04name\x18\x03 \x01(\tR\x04name*\x13\n\x04Type\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x32\x96\x08\n\x0cRayletDriver\x12\x35\n\x04Init\x12\x14.ray.rpc.InitRequest\x1a\x15.ray.rpc.InitResponse\"\x00\x12S\n\x0ePrepRuntimeEnv\x12\x1e.ray.rpc.PrepRuntimeEnvRequest\x1a\x1f.ray.rpc.PrepRuntimeEnvResponse\"\x00\x12:\n\tGetObject\x12\x13.ray.rpc.GetRequest\x1a\x14.ray.rpc.GetResponse\"\x00\x30\x01\x12\x38\n\tPutObject\x12\x13.ray.rpc.PutRequest\x1a\x14.ray.rpc.PutResponse\"\x00\x12;\n\nWaitObject\x12\x14.ray.rpc.WaitRequest\x1a\x15.ray.rpc.WaitResponse\"\x00\x12<\n\x08Schedule\x12\x13.ray.rpc.ClientTask\x1a\x19.ray.rpc.ClientTaskTicket\"\x00\x12\x44\n\tTerminate\x12\x19.ray.rpc.TerminateRequest\x1a\x1a.ray.rpc.TerminateResponse\"\x00\x12J\n\x0b\x43lusterInfo\x12\x1b.ray.rpc.ClusterInfoRequest\x1a\x1c.ray.rpc.ClusterInfoResponse\"\x00\x12\x38\n\x05KVGet\x12\x15.ray.rpc.KVGetRequest\x1a\x16.ray.rpc.KVGetResponse\"\x00\x12\x38\n\x05KVPut\x12\x15.ray.rpc.KVPutRequest\x1a\x16.ray.rpc.KVPutResponse\"\x00\x12\x38\n\x05KVDel\x12\x15.ray.rpc.KVDelRequest\x1a\x16.ray.rpc.KVDelResponse\"\x00\x12;\n\x06KVList\x12\x16.ray.rpc.KVListRequest\x1a\x17.ray.rpc.KVListResponse\"\x00\x12\x41\n\x08KVExists\x12\x18.ray.rpc.KVExistsRequest\x1a\x19.ray.rpc.KVExistsResponse\"\x00\x12\x62\n\x0fListNamedActors\x12%.ray.rpc.ClientListNamedActorsRequest\x1a&.ray.rpc.ClientListNamedActorsResponse\"\x00\x12\x65\n\x10PinRuntimeEnvURI\x12&.ray.rpc.ClientPinRuntimeEnvURIRequest\x1a\'.ray.rpc.ClientPinRuntimeEnvURIResponse\"\x00\x32S\n\x12RayletDataStreamer\x12=\n\x08\x44\x61tapath\x12\x14.ray.rpc.DataRequest\x1a\x15.ray.rpc.DataResponse\"\x00(\x01\x30\x01\x32U\n\x11RayletLogStreamer\x12@\n\tLogstream\x12\x1b.ray.rpc.LogSettingsRequest\x1a\x10.ray.rpc.LogData\"\x00(\x01\x30\x01\x42\x03\xf8\x01\x01\x62\x06proto3')

_TYPE = DESCRIPTOR.enum_types_by_name['Type']
Type = enum_type_wrapper.EnumTypeWrapper(_TYPE)
DEFAULT = 0


_ARG = DESCRIPTOR.message_types_by_name['Arg']
_TASKOPTIONS = DESCRIPTOR.message_types_by_name['TaskOptions']
_CLIENTTASK = DESCRIPTOR.message_types_by_name['ClientTask']
_CLIENTTASK_KWARGSENTRY = _CLIENTTASK.nested_types_by_name['KwargsEntry']
_CLIENTTASKTICKET = DESCRIPTOR.message_types_by_name['ClientTaskTicket']
_PUTREQUEST = DESCRIPTOR.message_types_by_name['PutRequest']
_PUTRESPONSE = DESCRIPTOR.message_types_by_name['PutResponse']
_GETREQUEST = DESCRIPTOR.message_types_by_name['GetRequest']
_GETRESPONSE = DESCRIPTOR.message_types_by_name['GetResponse']
_WAITREQUEST = DESCRIPTOR.message_types_by_name['WaitRequest']
_WAITRESPONSE = DESCRIPTOR.message_types_by_name['WaitResponse']
_CLUSTERINFOTYPE = DESCRIPTOR.message_types_by_name['ClusterInfoType']
_CLUSTERINFOREQUEST = DESCRIPTOR.message_types_by_name['ClusterInfoRequest']
_CLUSTERINFORESPONSE = DESCRIPTOR.message_types_by_name['ClusterInfoResponse']
_CLUSTERINFORESPONSE_RESOURCETABLE = _CLUSTERINFORESPONSE.nested_types_by_name['ResourceTable']
_CLUSTERINFORESPONSE_RESOURCETABLE_TABLEENTRY = _CLUSTERINFORESPONSE_RESOURCETABLE.nested_types_by_name['TableEntry']
_CLUSTERINFORESPONSE_RUNTIMECONTEXT = _CLUSTERINFORESPONSE.nested_types_by_name['RuntimeContext']
_TERMINATEREQUEST = DESCRIPTOR.message_types_by_name['TerminateRequest']
_TERMINATEREQUEST_ACTORTERMINATE = _TERMINATEREQUEST.nested_types_by_name['ActorTerminate']
_TERMINATEREQUEST_TASKOBJECTTERMINATE = _TERMINATEREQUEST.nested_types_by_name['TaskObjectTerminate']
_TERMINATERESPONSE = DESCRIPTOR.message_types_by_name['TerminateResponse']
_KVEXISTSREQUEST = DESCRIPTOR.message_types_by_name['KVExistsRequest']
_KVEXISTSRESPONSE = DESCRIPTOR.message_types_by_name['KVExistsResponse']
_KVGETREQUEST = DESCRIPTOR.message_types_by_name['KVGetRequest']
_KVGETRESPONSE = DESCRIPTOR.message_types_by_name['KVGetResponse']
_KVPUTREQUEST = DESCRIPTOR.message_types_by_name['KVPutRequest']
_KVPUTRESPONSE = DESCRIPTOR.message_types_by_name['KVPutResponse']
_KVDELREQUEST = DESCRIPTOR.message_types_by_name['KVDelRequest']
_KVDELRESPONSE = DESCRIPTOR.message_types_by_name['KVDelResponse']
_KVLISTREQUEST = DESCRIPTOR.message_types_by_name['KVListRequest']
_KVLISTRESPONSE = DESCRIPTOR.message_types_by_name['KVListResponse']
_CLIENTPINRUNTIMEENVURIREQUEST = DESCRIPTOR.message_types_by_name['ClientPinRuntimeEnvURIRequest']
_CLIENTPINRUNTIMEENVURIRESPONSE = DESCRIPTOR.message_types_by_name['ClientPinRuntimeEnvURIResponse']
_INITREQUEST = DESCRIPTOR.message_types_by_name['InitRequest']
_INITRESPONSE = DESCRIPTOR.message_types_by_name['InitResponse']
_PREPRUNTIMEENVREQUEST = DESCRIPTOR.message_types_by_name['PrepRuntimeEnvRequest']
_PREPRUNTIMEENVRESPONSE = DESCRIPTOR.message_types_by_name['PrepRuntimeEnvResponse']
_CLIENTLISTNAMEDACTORSREQUEST = DESCRIPTOR.message_types_by_name['ClientListNamedActorsRequest']
_CLIENTLISTNAMEDACTORSRESPONSE = DESCRIPTOR.message_types_by_name['ClientListNamedActorsResponse']
_RELEASEREQUEST = DESCRIPTOR.message_types_by_name['ReleaseRequest']
_RELEASERESPONSE = DESCRIPTOR.message_types_by_name['ReleaseResponse']
_CONNECTIONINFOREQUEST = DESCRIPTOR.message_types_by_name['ConnectionInfoRequest']
_CONNECTIONINFORESPONSE = DESCRIPTOR.message_types_by_name['ConnectionInfoResponse']
_CONNECTIONCLEANUPREQUEST = DESCRIPTOR.message_types_by_name['ConnectionCleanupRequest']
_CONNECTIONCLEANUPRESPONSE = DESCRIPTOR.message_types_by_name['ConnectionCleanupResponse']
_ACKNOWLEDGEREQUEST = DESCRIPTOR.message_types_by_name['AcknowledgeRequest']
_DATAREQUEST = DESCRIPTOR.message_types_by_name['DataRequest']
_DATARESPONSE = DESCRIPTOR.message_types_by_name['DataResponse']
_LOGSETTINGSREQUEST = DESCRIPTOR.message_types_by_name['LogSettingsRequest']
_LOGDATA = DESCRIPTOR.message_types_by_name['LogData']
_ARG_LOCALITY = _ARG.enum_types_by_name['Locality']
_CLIENTTASK_REMOTEEXECTYPE = _CLIENTTASK.enum_types_by_name['RemoteExecType']
_CLUSTERINFOTYPE_TYPEENUM = _CLUSTERINFOTYPE.enum_types_by_name['TypeEnum']
Arg = _reflection.GeneratedProtocolMessageType('Arg', (_message.Message,), {
  'DESCRIPTOR' : _ARG,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Arg)
  })
_sym_db.RegisterMessage(Arg)

TaskOptions = _reflection.GeneratedProtocolMessageType('TaskOptions', (_message.Message,), {
  'DESCRIPTOR' : _TASKOPTIONS,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskOptions)
  })
_sym_db.RegisterMessage(TaskOptions)

ClientTask = _reflection.GeneratedProtocolMessageType('ClientTask', (_message.Message,), {

  'KwargsEntry' : _reflection.GeneratedProtocolMessageType('KwargsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CLIENTTASK_KWARGSENTRY,
    '__module__' : 'src.ray.protobuf.ray_client_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ClientTask.KwargsEntry)
    })
  ,
  'DESCRIPTOR' : _CLIENTTASK,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ClientTask)
  })
_sym_db.RegisterMessage(ClientTask)
_sym_db.RegisterMessage(ClientTask.KwargsEntry)

ClientTaskTicket = _reflection.GeneratedProtocolMessageType('ClientTaskTicket', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTTASKTICKET,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ClientTaskTicket)
  })
_sym_db.RegisterMessage(ClientTaskTicket)

PutRequest = _reflection.GeneratedProtocolMessageType('PutRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUTREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PutRequest)
  })
_sym_db.RegisterMessage(PutRequest)

PutResponse = _reflection.GeneratedProtocolMessageType('PutResponse', (_message.Message,), {
  'DESCRIPTOR' : _PUTRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PutResponse)
  })
_sym_db.RegisterMessage(PutResponse)

GetRequest = _reflection.GeneratedProtocolMessageType('GetRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetRequest)
  })
_sym_db.RegisterMessage(GetRequest)

GetResponse = _reflection.GeneratedProtocolMessageType('GetResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetResponse)
  })
_sym_db.RegisterMessage(GetResponse)

WaitRequest = _reflection.GeneratedProtocolMessageType('WaitRequest', (_message.Message,), {
  'DESCRIPTOR' : _WAITREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WaitRequest)
  })
_sym_db.RegisterMessage(WaitRequest)

WaitResponse = _reflection.GeneratedProtocolMessageType('WaitResponse', (_message.Message,), {
  'DESCRIPTOR' : _WAITRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WaitResponse)
  })
_sym_db.RegisterMessage(WaitResponse)

ClusterInfoType = _reflection.GeneratedProtocolMessageType('ClusterInfoType', (_message.Message,), {
  'DESCRIPTOR' : _CLUSTERINFOTYPE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ClusterInfoType)
  })
_sym_db.RegisterMessage(ClusterInfoType)

ClusterInfoRequest = _reflection.GeneratedProtocolMessageType('ClusterInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLUSTERINFOREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ClusterInfoRequest)
  })
_sym_db.RegisterMessage(ClusterInfoRequest)

ClusterInfoResponse = _reflection.GeneratedProtocolMessageType('ClusterInfoResponse', (_message.Message,), {

  'ResourceTable' : _reflection.GeneratedProtocolMessageType('ResourceTable', (_message.Message,), {

    'TableEntry' : _reflection.GeneratedProtocolMessageType('TableEntry', (_message.Message,), {
      'DESCRIPTOR' : _CLUSTERINFORESPONSE_RESOURCETABLE_TABLEENTRY,
      '__module__' : 'src.ray.protobuf.ray_client_pb2'
      # @@protoc_insertion_point(class_scope:ray.rpc.ClusterInfoResponse.ResourceTable.TableEntry)
      })
    ,
    'DESCRIPTOR' : _CLUSTERINFORESPONSE_RESOURCETABLE,
    '__module__' : 'src.ray.protobuf.ray_client_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ClusterInfoResponse.ResourceTable)
    })
  ,

  'RuntimeContext' : _reflection.GeneratedProtocolMessageType('RuntimeContext', (_message.Message,), {
    'DESCRIPTOR' : _CLUSTERINFORESPONSE_RUNTIMECONTEXT,
    '__module__' : 'src.ray.protobuf.ray_client_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ClusterInfoResponse.RuntimeContext)
    })
  ,
  'DESCRIPTOR' : _CLUSTERINFORESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ClusterInfoResponse)
  })
_sym_db.RegisterMessage(ClusterInfoResponse)
_sym_db.RegisterMessage(ClusterInfoResponse.ResourceTable)
_sym_db.RegisterMessage(ClusterInfoResponse.ResourceTable.TableEntry)
_sym_db.RegisterMessage(ClusterInfoResponse.RuntimeContext)

TerminateRequest = _reflection.GeneratedProtocolMessageType('TerminateRequest', (_message.Message,), {

  'ActorTerminate' : _reflection.GeneratedProtocolMessageType('ActorTerminate', (_message.Message,), {
    'DESCRIPTOR' : _TERMINATEREQUEST_ACTORTERMINATE,
    '__module__' : 'src.ray.protobuf.ray_client_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.TerminateRequest.ActorTerminate)
    })
  ,

  'TaskObjectTerminate' : _reflection.GeneratedProtocolMessageType('TaskObjectTerminate', (_message.Message,), {
    'DESCRIPTOR' : _TERMINATEREQUEST_TASKOBJECTTERMINATE,
    '__module__' : 'src.ray.protobuf.ray_client_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.TerminateRequest.TaskObjectTerminate)
    })
  ,
  'DESCRIPTOR' : _TERMINATEREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TerminateRequest)
  })
_sym_db.RegisterMessage(TerminateRequest)
_sym_db.RegisterMessage(TerminateRequest.ActorTerminate)
_sym_db.RegisterMessage(TerminateRequest.TaskObjectTerminate)

TerminateResponse = _reflection.GeneratedProtocolMessageType('TerminateResponse', (_message.Message,), {
  'DESCRIPTOR' : _TERMINATERESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TerminateResponse)
  })
_sym_db.RegisterMessage(TerminateResponse)

KVExistsRequest = _reflection.GeneratedProtocolMessageType('KVExistsRequest', (_message.Message,), {
  'DESCRIPTOR' : _KVEXISTSREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVExistsRequest)
  })
_sym_db.RegisterMessage(KVExistsRequest)

KVExistsResponse = _reflection.GeneratedProtocolMessageType('KVExistsResponse', (_message.Message,), {
  'DESCRIPTOR' : _KVEXISTSRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVExistsResponse)
  })
_sym_db.RegisterMessage(KVExistsResponse)

KVGetRequest = _reflection.GeneratedProtocolMessageType('KVGetRequest', (_message.Message,), {
  'DESCRIPTOR' : _KVGETREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVGetRequest)
  })
_sym_db.RegisterMessage(KVGetRequest)

KVGetResponse = _reflection.GeneratedProtocolMessageType('KVGetResponse', (_message.Message,), {
  'DESCRIPTOR' : _KVGETRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVGetResponse)
  })
_sym_db.RegisterMessage(KVGetResponse)

KVPutRequest = _reflection.GeneratedProtocolMessageType('KVPutRequest', (_message.Message,), {
  'DESCRIPTOR' : _KVPUTREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVPutRequest)
  })
_sym_db.RegisterMessage(KVPutRequest)

KVPutResponse = _reflection.GeneratedProtocolMessageType('KVPutResponse', (_message.Message,), {
  'DESCRIPTOR' : _KVPUTRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVPutResponse)
  })
_sym_db.RegisterMessage(KVPutResponse)

KVDelRequest = _reflection.GeneratedProtocolMessageType('KVDelRequest', (_message.Message,), {
  'DESCRIPTOR' : _KVDELREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVDelRequest)
  })
_sym_db.RegisterMessage(KVDelRequest)

KVDelResponse = _reflection.GeneratedProtocolMessageType('KVDelResponse', (_message.Message,), {
  'DESCRIPTOR' : _KVDELRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVDelResponse)
  })
_sym_db.RegisterMessage(KVDelResponse)

KVListRequest = _reflection.GeneratedProtocolMessageType('KVListRequest', (_message.Message,), {
  'DESCRIPTOR' : _KVLISTREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVListRequest)
  })
_sym_db.RegisterMessage(KVListRequest)

KVListResponse = _reflection.GeneratedProtocolMessageType('KVListResponse', (_message.Message,), {
  'DESCRIPTOR' : _KVLISTRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.KVListResponse)
  })
_sym_db.RegisterMessage(KVListResponse)

ClientPinRuntimeEnvURIRequest = _reflection.GeneratedProtocolMessageType('ClientPinRuntimeEnvURIRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTPINRUNTIMEENVURIREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ClientPinRuntimeEnvURIRequest)
  })
_sym_db.RegisterMessage(ClientPinRuntimeEnvURIRequest)

ClientPinRuntimeEnvURIResponse = _reflection.GeneratedProtocolMessageType('ClientPinRuntimeEnvURIResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTPINRUNTIMEENVURIRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ClientPinRuntimeEnvURIResponse)
  })
_sym_db.RegisterMessage(ClientPinRuntimeEnvURIResponse)

InitRequest = _reflection.GeneratedProtocolMessageType('InitRequest', (_message.Message,), {
  'DESCRIPTOR' : _INITREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.InitRequest)
  })
_sym_db.RegisterMessage(InitRequest)

InitResponse = _reflection.GeneratedProtocolMessageType('InitResponse', (_message.Message,), {
  'DESCRIPTOR' : _INITRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.InitResponse)
  })
_sym_db.RegisterMessage(InitResponse)

PrepRuntimeEnvRequest = _reflection.GeneratedProtocolMessageType('PrepRuntimeEnvRequest', (_message.Message,), {
  'DESCRIPTOR' : _PREPRUNTIMEENVREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PrepRuntimeEnvRequest)
  })
_sym_db.RegisterMessage(PrepRuntimeEnvRequest)

PrepRuntimeEnvResponse = _reflection.GeneratedProtocolMessageType('PrepRuntimeEnvResponse', (_message.Message,), {
  'DESCRIPTOR' : _PREPRUNTIMEENVRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PrepRuntimeEnvResponse)
  })
_sym_db.RegisterMessage(PrepRuntimeEnvResponse)

ClientListNamedActorsRequest = _reflection.GeneratedProtocolMessageType('ClientListNamedActorsRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTLISTNAMEDACTORSREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ClientListNamedActorsRequest)
  })
_sym_db.RegisterMessage(ClientListNamedActorsRequest)

ClientListNamedActorsResponse = _reflection.GeneratedProtocolMessageType('ClientListNamedActorsResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTLISTNAMEDACTORSRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ClientListNamedActorsResponse)
  })
_sym_db.RegisterMessage(ClientListNamedActorsResponse)

ReleaseRequest = _reflection.GeneratedProtocolMessageType('ReleaseRequest', (_message.Message,), {
  'DESCRIPTOR' : _RELEASEREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReleaseRequest)
  })
_sym_db.RegisterMessage(ReleaseRequest)

ReleaseResponse = _reflection.GeneratedProtocolMessageType('ReleaseResponse', (_message.Message,), {
  'DESCRIPTOR' : _RELEASERESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReleaseResponse)
  })
_sym_db.RegisterMessage(ReleaseResponse)

ConnectionInfoRequest = _reflection.GeneratedProtocolMessageType('ConnectionInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTIONINFOREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ConnectionInfoRequest)
  })
_sym_db.RegisterMessage(ConnectionInfoRequest)

ConnectionInfoResponse = _reflection.GeneratedProtocolMessageType('ConnectionInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTIONINFORESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ConnectionInfoResponse)
  })
_sym_db.RegisterMessage(ConnectionInfoResponse)

ConnectionCleanupRequest = _reflection.GeneratedProtocolMessageType('ConnectionCleanupRequest', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTIONCLEANUPREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ConnectionCleanupRequest)
  })
_sym_db.RegisterMessage(ConnectionCleanupRequest)

ConnectionCleanupResponse = _reflection.GeneratedProtocolMessageType('ConnectionCleanupResponse', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTIONCLEANUPRESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ConnectionCleanupResponse)
  })
_sym_db.RegisterMessage(ConnectionCleanupResponse)

AcknowledgeRequest = _reflection.GeneratedProtocolMessageType('AcknowledgeRequest', (_message.Message,), {
  'DESCRIPTOR' : _ACKNOWLEDGEREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.AcknowledgeRequest)
  })
_sym_db.RegisterMessage(AcknowledgeRequest)

DataRequest = _reflection.GeneratedProtocolMessageType('DataRequest', (_message.Message,), {
  'DESCRIPTOR' : _DATAREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.DataRequest)
  })
_sym_db.RegisterMessage(DataRequest)

DataResponse = _reflection.GeneratedProtocolMessageType('DataResponse', (_message.Message,), {
  'DESCRIPTOR' : _DATARESPONSE,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.DataResponse)
  })
_sym_db.RegisterMessage(DataResponse)

LogSettingsRequest = _reflection.GeneratedProtocolMessageType('LogSettingsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGSETTINGSREQUEST,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.LogSettingsRequest)
  })
_sym_db.RegisterMessage(LogSettingsRequest)

LogData = _reflection.GeneratedProtocolMessageType('LogData', (_message.Message,), {
  'DESCRIPTOR' : _LOGDATA,
  '__module__' : 'src.ray.protobuf.ray_client_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.LogData)
  })
_sym_db.RegisterMessage(LogData)

_RAYLETDRIVER = DESCRIPTOR.services_by_name['RayletDriver']
_RAYLETDATASTREAMER = DESCRIPTOR.services_by_name['RayletDataStreamer']
_RAYLETLOGSTREAMER = DESCRIPTOR.services_by_name['RayletLogStreamer']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _CLIENTTASK_KWARGSENTRY._options = None
  _CLIENTTASK_KWARGSENTRY._serialized_options = b'8\001'
  _GETREQUEST.fields_by_name['id']._options = None
  _GETREQUEST.fields_by_name['id']._serialized_options = b'\030\001'
  _CLUSTERINFORESPONSE_RESOURCETABLE_TABLEENTRY._options = None
  _CLUSTERINFORESPONSE_RESOURCETABLE_TABLEENTRY._serialized_options = b'8\001'
  _TYPE._serialized_start=6404
  _TYPE._serialized_end=6423
  _ARG._serialized_start=47
  _ARG._serialized_end=228
  _ARG_LOCALITY._serialized_start=189
  _ARG_LOCALITY._serialized_end=228
  _TASKOPTIONS._serialized_start=230
  _TASKOPTIONS._serialized_end=284
  _CLIENTTASK._serialized_start=287
  _CLIENTTASK._serialized_end=915
  _CLIENTTASK_KWARGSENTRY._serialized_start=753
  _CLIENTTASK_KWARGSENTRY._serialized_end=824
  _CLIENTTASK_REMOTEEXECTYPE._serialized_start=826
  _CLIENTTASK_REMOTEEXECTYPE._serialized_end=915
  _CLIENTTASKTICKET._serialized_start=917
  _CLIENTTASKTICKET._serialized_end=1010
  _PUTREQUEST._serialized_start=1013
  _PUTREQUEST._serialized_end=1201
  _PUTRESPONSE._serialized_start=1203
  _PUTRESPONSE._serialized_end=1276
  _GETREQUEST._serialized_start=1279
  _GETREQUEST._serialized_end=1429
  _GETRESPONSE._serialized_start=1432
  _GETRESPONSE._serialized_end=1602
  _WAITREQUEST._serialized_start=1605
  _WAITREQUEST._serialized_end=1737
  _WAITRESPONSE._serialized_start=1740
  _WAITRESPONSE._serialized_end=1868
  _CLUSTERINFOTYPE._serialized_start=1871
  _CLUSTERINFOTYPE._serialized_end=2044
  _CLUSTERINFOTYPE_TYPEENUM._serialized_start=1891
  _CLUSTERINFOTYPE_TYPEENUM._serialized_end=2044
  _CLUSTERINFOREQUEST._serialized_start=2046
  _CLUSTERINFOREQUEST._serialized_end=2121
  _CLUSTERINFORESPONSE._serialized_start=2124
  _CLUSTERINFORESPONSE._serialized_end=2778
  _CLUSTERINFORESPONSE_RESOURCETABLE._serialized_start=2398
  _CLUSTERINFORESPONSE_RESOURCETABLE._serialized_end=2548
  _CLUSTERINFORESPONSE_RESOURCETABLE_TABLEENTRY._serialized_start=2492
  _CLUSTERINFORESPONSE_RESOURCETABLE_TABLEENTRY._serialized_end=2548
  _CLUSTERINFORESPONSE_RUNTIMECONTEXT._serialized_start=2551
  _CLUSTERINFORESPONSE_RUNTIMECONTEXT._serialized_end=2761
  _TERMINATEREQUEST._serialized_start=2781
  _TERMINATEREQUEST._serialized_end=3150
  _TERMINATEREQUEST_ACTORTERMINATE._serialized_start=2978
  _TERMINATEREQUEST_ACTORTERMINATE._serialized_end=3041
  _TERMINATEREQUEST_TASKOBJECTTERMINATE._serialized_start=3043
  _TERMINATEREQUEST_TASKOBJECTTERMINATE._serialized_end=3132
  _TERMINATERESPONSE._serialized_start=3152
  _TERMINATERESPONSE._serialized_end=3187
  _KVEXISTSREQUEST._serialized_start=3189
  _KVEXISTSREQUEST._serialized_end=3273
  _KVEXISTSRESPONSE._serialized_start=3275
  _KVEXISTSRESPONSE._serialized_end=3317
  _KVGETREQUEST._serialized_start=3319
  _KVGETREQUEST._serialized_end=3400
  _KVGETRESPONSE._serialized_start=3402
  _KVGETRESPONSE._serialized_end=3454
  _KVPUTREQUEST._serialized_start=3457
  _KVPUTREQUEST._serialized_end=3590
  _KVPUTRESPONSE._serialized_start=3592
  _KVPUTRESPONSE._serialized_end=3646
  _KVDELREQUEST._serialized_start=3648
  _KVDELREQUEST._serialized_end=3765
  _KVDELRESPONSE._serialized_start=3767
  _KVDELRESPONSE._serialized_end=3815
  _KVLISTREQUEST._serialized_start=3817
  _KVLISTREQUEST._serialized_end=3905
  _KVLISTRESPONSE._serialized_start=3907
  _KVLISTRESPONSE._serialized_end=3943
  _CLIENTPINRUNTIMEENVURIREQUEST._serialized_start=3945
  _CLIENTPINRUNTIMEENVURIREQUEST._serialized_end=4029
  _CLIENTPINRUNTIMEENVURIRESPONSE._serialized_start=4031
  _CLIENTPINRUNTIMEENVURIRESPONSE._serialized_end=4063
  _INITREQUEST._serialized_start=4066
  _INITREQUEST._serialized_end=4204
  _INITRESPONSE._serialized_start=4206
  _INITRESPONSE._serialized_end=4254
  _PREPRUNTIMEENVREQUEST._serialized_start=4256
  _PREPRUNTIMEENVREQUEST._serialized_end=4279
  _PREPRUNTIMEENVRESPONSE._serialized_start=4281
  _PREPRUNTIMEENVRESPONSE._serialized_end=4305
  _CLIENTLISTNAMEDACTORSREQUEST._serialized_start=4307
  _CLIENTLISTNAMEDACTORSREQUEST._serialized_end=4376
  _CLIENTLISTNAMEDACTORSRESPONSE._serialized_start=4378
  _CLIENTLISTNAMEDACTORSRESPONSE._serialized_end=4442
  _RELEASEREQUEST._serialized_start=4444
  _RELEASEREQUEST._serialized_end=4478
  _RELEASERESPONSE._serialized_start=4480
  _RELEASERESPONSE._serialized_end=4513
  _CONNECTIONINFOREQUEST._serialized_start=4515
  _CONNECTIONINFOREQUEST._serialized_end=4538
  _CONNECTIONINFORESPONSE._serialized_start=4541
  _CONNECTIONINFORESPONSE._serialized_end=4744
  _CONNECTIONCLEANUPREQUEST._serialized_start=4746
  _CONNECTIONCLEANUPREQUEST._serialized_end=4772
  _CONNECTIONCLEANUPRESPONSE._serialized_start=4774
  _CONNECTIONCLEANUPRESPONSE._serialized_end=4801
  _ACKNOWLEDGEREQUEST._serialized_start=4803
  _ACKNOWLEDGEREQUEST._serialized_end=4846
  _DATAREQUEST._serialized_start=4849
  _DATAREQUEST._serialized_end=5559
  _DATARESPONSE._serialized_start=5562
  _DATARESPONSE._serialized_end=6255
  _LOGSETTINGSREQUEST._serialized_start=6257
  _LOGSETTINGSREQUEST._serialized_end=6331
  _LOGDATA._serialized_start=6333
  _LOGDATA._serialized_end=6402
  _RAYLETDRIVER._serialized_start=6426
  _RAYLETDRIVER._serialized_end=7472
  _RAYLETDATASTREAMER._serialized_start=7474
  _RAYLETDATASTREAMER._serialized_end=7557
  _RAYLETLOGSTREAMER._serialized_start=7559
  _RAYLETLOGSTREAMER._serialized_end=7644
# @@protoc_insertion_point(module_scope)



import warnings
from typing import Dict, Optional
from ray.air.execution.resources.request import ResourceRequest
from ray.util.annotations import DeveloperAPI, PublicAPI
from ray.util.placement_group import placement_group


@PublicAPI(stability="beta")
class PlacementGroupFactory(ResourceRequest):
    """Wrapper class that creates placement groups for trials.

    This function should be used to define resource requests for Ray Tune
    trials. It holds the parameters to create
    :ref:`placement groups <ray-placement-group-doc-ref>`.
    At a minimum, this will hold at least one bundle specifying the
    resource requirements for each trial:

    .. code-block:: python

        from ray import tune

        tuner = tune.Tuner(
            tune.with_resources(
                train,
                resources=tune.PlacementGroupFactory([
                    {"CPU": 1, "GPU": 0.5, "custom_resource": 2}
                ])
            )
        )
        tuner.fit()

    If the trial itself schedules further remote workers, the resource
    requirements should be specified in additional bundles. You can also
    pass the placement strategy for these bundles, e.g. to enforce
    co-located placement:

    .. code-block:: python

        from ray import tune

        tuner = tune.Tuner(
            tune.with_resources(
                train,
                resources=tune.PlacementGroupFactory([
                    {"CPU": 1, "GPU": 0.5, "custom_resource": 2},
                    {"CPU": 2},
                    {"CPU": 2},
                ], strategy="PACK")
            )
        )
        tuner.fit()

    The example above will reserve 1 CPU, 0.5 GPUs and 2 custom_resources
    for the trainable itself, and reserve another 2 bundles of 2 CPUs each.
    The trial will only start when all these resources are available. This
    could be used e.g. if you had one learner running in the main trainable
    that schedules two remote workers that need access to 2 CPUs each.

    If the trainable itself doesn't require resources.
    You can specify it as:

    .. code-block:: python

        from ray import tune

        tuner = tune.Tuner(
            tune.with_resources(
                train,
                resources=tune.PlacementGroupFactory([
                    {},
                    {"CPU": 2},
                    {"CPU": 2},
                ], strategy="PACK")
            )
        )
        tuner.fit()

    Args:
        bundles: A list of bundles which
            represent the resources requirements.
        strategy: The strategy to create the placement group.

         - "PACK": Packs Bundles into as few nodes as possible.
         - "SPREAD": Places Bundles across distinct nodes as even as possible.
         - "STRICT_PACK": Packs Bundles into one node. The group is
           not allowed to span multiple nodes.
         - "STRICT_SPREAD": Packs Bundles across distinct nodes.
        *args: Passed to the call of ``placement_group()``
        **kwargs: Passed to the call of ``placement_group()``

    """

    def __call__(self, *args, **kwargs):
        warnings.warn(
            "Calling PlacementGroupFactory objects is deprecated. Use "
            "`to_placement_group()` instead.",
            DeprecationWarning,
        )
        kwargs.update(self._bound.kwargs)
        # Call with bounded *args and **kwargs
        return placement_group(*self._bound.args, **kwargs)


@DeveloperAPI
def resource_dict_to_pg_factory(spec: Optional[Dict[str, float]] = None):
    """Translates resource dict into PlacementGroupFactory."""
    spec = spec or {"cpu": 1}

    spec = spec.copy()

    cpus = spec.pop("cpu", spec.pop("CPU", 0.0))
    gpus = spec.pop("gpu", spec.pop("GPU", 0.0))
    memory = spec.pop("memory", 0.0)

    # If there is a custom_resources key, use as base for bundle
    bundle = {k: v for k, v in spec.pop("custom_resources", {}).items()}

    # Otherwise, consider all other keys as custom resources
    if not bundle:
        bundle = spec

    bundle.update(
        {
            "CPU": cpus,
            "GPU": gpus,
            "memory": memory,
        }
    )

    return PlacementGroupFactory([bundle])



import logging
import numpy as np
import math
import pandas as pd

from typing import Dict, Any, Optional, List

from ray.data import Dataset

from ray.rllib.policy import Policy
from ray.rllib.policy.sample_batch import SampleBatch, convert_ma_batch_to_sample_batch
from ray.rllib.utils.annotations import DeveloperAPI, override
from ray.rllib.utils.typing import SampleBatchType
from ray.rllib.utils.numpy import convert_to_numpy

from ray.rllib.offline.estimators.off_policy_estimator import OffPolicyEstimator
from ray.rllib.offline.estimators.fqe_torch_model import FQETorchModel
from ray.rllib.offline.offline_evaluator import OfflineEvaluator
from ray.rllib.offline.offline_evaluation_utils import (
    compute_is_weights,
    compute_q_and_v_values,
)

logger = logging.getLogger()


@DeveloperAPI
class DoublyRobust(OffPolicyEstimator):
    """The Doubly Robust estimator.

    Let s_t, a_t, and r_t be the state, action, and reward at timestep t.

    This method trains a Q-model for the evaluation policy \pi_e on behavior
    data generated by \pi_b. Currently, RLlib implements this using
    Fitted-Q Evaluation (FQE). You can also implement your own model
    and pass it in as `q_model_config = {"type": your_model_class, **your_kwargs}`.

    For behavior policy \pi_b and evaluation policy \pi_e, define the
    cumulative importance ratio at timestep t as:
    p_t = \sum_{t'=0}^t (\pi_e(a_{t'} | s_{t'}) / \pi_b(a_{t'} | s_{t'})).

    Consider an episode with length T. Let V_T = 0.
    For all t in {0, T - 1}, use the following recursive update:
    V_t^DR = (\sum_{a \in A} \pi_e(a | s_t) Q(s_t, a))
        + p_t * (r_t + \gamma * V_{t+1}^DR - Q(s_t, a_t))

    This estimator computes the expected return for \pi_e for an episode as:
    V^{\pi_e}(s_0) = V_0^DR
    and returns the mean and standard deviation over episodes.

    For more information refer to https://arxiv.org/pdf/1911.06854.pdf"""

    @override(OffPolicyEstimator)
    def __init__(
        self,
        policy: Policy,
        gamma: float,
        epsilon_greedy: float = 0.0,
        normalize_weights: bool = True,
        q_model_config: Optional[Dict] = None,
    ):
        """Initializes a Doubly Robust OPE Estimator.

        Args:
            policy: Policy to evaluate.
            gamma: Discount factor of the environment.
            epsilon_greedy: The probability by which we act acording to a fully random
                policy during deployment. With 1-epsilon_greedy we act
                according the target policy.
            normalize_weights: If True, the inverse propensity scores are normalized to
                their sum across the entire dataset. The effect of this is similar to
                weighted importance sampling compared to standard importance sampling.
            q_model_config: Arguments to specify the Q-model. Must specify
                a `type` key pointing to the Q-model class.
                This Q-model is trained in the train() method and is used
                to compute the state-value and Q-value estimates
                for the DoublyRobust estimator.
                It must implement `train`, `estimate_q`, and `estimate_v`.
                TODO (Rohan138): Unify this with RLModule API.
        """

        super().__init__(policy, gamma, epsilon_greedy)
        q_model_config = q_model_config or {}
        q_model_config["gamma"] = gamma

        self._model_cls = q_model_config.pop("type", FQETorchModel)
        self._model_configs = q_model_config
        self._normalize_weights = normalize_weights

        self.model = self._model_cls(
            policy=policy,
            **q_model_config,
        )
        assert hasattr(
            self.model, "estimate_v"
        ), "self.model must implement `estimate_v`!"
        assert hasattr(
            self.model, "estimate_q"
        ), "self.model must implement `estimate_q`!"

    @override(OffPolicyEstimator)
    def estimate_on_single_episode(self, episode: SampleBatch) -> Dict[str, Any]:
        estimates_per_epsiode = {}

        rewards, old_prob = episode["rewards"], episode["action_prob"]
        new_prob = self.compute_action_probs(episode)

        weight = new_prob / old_prob

        v_behavior = 0.0
        v_target = 0.0
        q_values = self.model.estimate_q(episode)
        q_values = convert_to_numpy(q_values)
        v_values = self.model.estimate_v(episode)
        v_values = convert_to_numpy(v_values)
        assert q_values.shape == v_values.shape == (episode.count,)

        for t in reversed(range(episode.count)):
            v_behavior = rewards[t] + self.gamma * v_behavior
            v_target = v_values[t] + weight[t] * (
                rewards[t] + self.gamma * v_target - q_values[t]
            )
        v_target = v_target.item()

        estimates_per_epsiode["v_behavior"] = v_behavior
        estimates_per_epsiode["v_target"] = v_target

        return estimates_per_epsiode

    @override(OffPolicyEstimator)
    def estimate_on_single_step_samples(
        self, batch: SampleBatch
    ) -> Dict[str, List[float]]:
        estimates_per_epsiode = {}

        rewards, old_prob = batch["rewards"], batch["action_prob"]
        new_prob = self.compute_action_probs(batch)

        q_values = self.model.estimate_q(batch)
        q_values = convert_to_numpy(q_values)
        v_values = self.model.estimate_v(batch)
        v_values = convert_to_numpy(v_values)

        v_behavior = rewards

        weight = new_prob / old_prob
        v_target = v_values + weight * (rewards - q_values)

        estimates_per_epsiode["v_behavior"] = v_behavior
        estimates_per_epsiode["v_target"] = v_target

        return estimates_per_epsiode

    @override(OffPolicyEstimator)
    def train(self, batch: SampleBatchType) -> Dict[str, Any]:
        """Trains self.model on the given batch.

        Args:
        batch: A SampleBatch or MultiAgentbatch to train on

        Returns:
            A dict with key "loss" and value as the mean training loss.
        """
        batch = convert_ma_batch_to_sample_batch(batch)
        losses = self.model.train(batch)
        return {"loss": np.mean(losses)}

    @override(OfflineEvaluator)
    def estimate_on_dataset(
        self, dataset: Dataset, *, n_parallelism: int = ...
    ) -> Dict[str, Any]:
        """Estimates the policy value using the Doubly Robust estimator.

        The doubly robust estimator uses normalization of importance sampling weights
        (aka. propensity ratios) to the average of the importance weights across the
        entire dataset. This is done to reduce the variance of the estimate (similar to
        weighted importance sampling). You can disable this by setting
        `normalize_weights=False` in the constructor.

        Note: This estimate works for only discrete action spaces for now.

        Args:
            dataset: Dataset to compute the estimate on. Each record in dataset should
                include the following columns: `obs`, `actions`, `action_prob` and
                `rewards`. The `obs` on each row shoud be a vector of D dimensions.
            n_parallelism: Number of parallelism to use for the computation.

        Returns:
            A dict with the following keys:
                v_target: The estimated value of the target policy.
                v_behavior: The estimated value of the behavior policy.
                v_gain: The estimated gain of the target policy over the behavior
                    policy.
                v_std: The standard deviation of the estimated value of the target.
        """

        # step 1: compute the weights and weighted rewards
        batch_size = max(dataset.count() // n_parallelism, 1)
        updated_ds = dataset.map_batches(
            compute_is_weights,
            batch_size=batch_size,
            batch_format="pandas",
            fn_kwargs={
                "policy_state": self.policy.get_state(),
                "estimator_class": self.__class__,
            },
        )

        # step 2: compute q_values and v_values
        batch_size = max(updated_ds.count() // n_parallelism, 1)
        updated_ds = updated_ds.map_batches(
            compute_q_and_v_values,
            batch_size=batch_size,
            batch_format="pandas",
            fn_kwargs={
                "model_class": self.model.__class__,
                "model_state": self.model.get_state(),
            },
        )

        # step 3: compute the v_target
        def compute_v_target(batch: pd.DataFrame, normalizer: float = 1.0):
            weights = batch["weights"] / normalizer
            batch["v_target"] = batch["v_values"] + weights * (
                batch["rewards"] - batch["q_values"]
            )
            batch["v_behavior"] = batch["rewards"]
            return batch

        normalizer = updated_ds.mean("weights") if self._normalize_weights else 1.0
        updated_ds = updated_ds.map_batches(
            compute_v_target,
            batch_size=batch_size,
            batch_format="pandas",
            fn_kwargs={"normalizer": normalizer},
        )

        v_behavior = updated_ds.mean("v_behavior")
        v_target = updated_ds.mean("v_target")
        v_gain_mean = v_target / v_behavior
        v_gain_ste = (
            updated_ds.std("v_target")
            / normalizer
            / v_behavior
            / math.sqrt(dataset.count())
        )

        return {
            "v_behavior": v_behavior,
            "v_target": v_target,
            "v_gain_mean": v_gain_mean,
            "v_gain_ste": v_gain_ste,
        }



"""This is the next version of action distribution base class."""
from typing import Tuple
import gymnasium as gym
import abc

from ray.rllib.utils.annotations import ExperimentalAPI
from ray.rllib.utils.typing import TensorType, Union
from ray.rllib.utils.annotations import override


@ExperimentalAPI
class Distribution(abc.ABC):
    """The base class for distribution over a random variable.

    Examples:
        >>> model = ... # a model that outputs a vector of logits
        >>> action_logits = model.forward(obs)
        >>> action_dist = Distribution(action_logits)
        >>> action = action_dist.sample()
        >>> logp = action_dist.logp(action)
        >>> kl = action_dist.kl(action_dist2)
        >>> entropy = action_dist.entropy()
    """

    @abc.abstractmethod
    def sample(
        self,
        *,
        sample_shape: Tuple[int, ...] = None,
        return_logp: bool = False,
        **kwargs,
    ) -> Union[TensorType, Tuple[TensorType, TensorType]]:
        """Draw a sample from the distribution.

        Args:
            sample_shape: The shape of the sample to draw.
            return_logp: Whether to return the logp of the sampled values.
            **kwargs: Forward compatibility placeholder.

        Returns:
            The sampled values. If return_logp is True, returns a tuple of the
            sampled values and its logp.
        """

    @abc.abstractmethod
    def rsample(
        self,
        *,
        sample_shape: Tuple[int, ...] = None,
        return_logp: bool = False,
        **kwargs,
    ) -> Union[TensorType, Tuple[TensorType, TensorType]]:
        """Draw a re-parameterized sample from the action distribution.

        If this method is implemented, we can take gradients of samples w.r.t. the
        distribution parameters.

        Args:
            sample_shape: The shape of the sample to draw.
            return_logp: Whether to return the logp of the sampled values.
            **kwargs: Forward compatibility placeholder.

        Returns:
            The sampled values. If return_logp is True, returns a tuple of the
            sampled values and its logp.
        """

    @abc.abstractmethod
    def logp(self, value: TensorType, **kwargs) -> TensorType:
        """The log-likelihood of the distribution computed at `value`

        Args:
            value: The value to compute the log-likelihood at.
            **kwargs: Forward compatibility placeholder.

        Returns:
            The log-likelihood of the value.
        """

    @abc.abstractmethod
    def kl(self, other: "Distribution", **kwargs) -> TensorType:
        """The KL-divergence between two distributions.

        Args:
            other: The other distribution.
            **kwargs: Forward compatibility placeholder.

        Returns:
            The KL-divergence between the two distributions.
        """

    @abc.abstractmethod
    def entropy(self, **kwargs) -> TensorType:
        """The entropy of the distribution.

        Args:
            **kwargs: Forward compatibility placeholder.

        Returns:
            The entropy of the distribution.
        """

    @staticmethod
    @abc.abstractmethod
    def required_input_dim(space: gym.Space, **kwargs) -> int:
        """Returns the required length of an input parameter tensor.

        Args:
            space: The space this distribution will be used for,
                whose shape attributes will be used to determine the required shape of
                the input parameter tensor.
            **kwargs: Forward compatibility placeholder.

        Returns:
            size of the required input vector (minus leading batch dimension).
        """

    @classmethod
    def from_logits(cls, logits: TensorType, **kwargs) -> "Distribution":
        """Creates a Distribution from logits.

        The caller does not need to have knowledge of the distribution class in order
        to create it and sample from it. The passed batched logits vectors might be
        split up and are passed to the distribution class' constructor as kwargs.

        Args:
            logits: The logits to create the distribution from.
            **kwargs: Forward compatibility placeholder.

        Returns:
            The created distribution.

        .. code-block:: python

            import numpy as np
            from ray.rllib.models.distributions import Distribution

            class Uniform(Distribution):
                def __init__(self, lower, upper):
                    self.lower = lower
                    self.upper = upper

                def sample(self):
                    return self.lower + (self.upper - self.lower) * np.random.rand()

                def logp(self, x):
                    ...

                def kl(self, other):
                    ...

                def entropy(self):
                    ...

                @staticmethod
                def required_input_dim(space):
                    ...

                def rsample(self):
                    ...

                @classmethod
                def from_logits(cls, logits, **kwargs):
                    return Uniform(logits[:, 0], logits[:, 1])

            logits = np.array([[0.0, 1.0], [2.0, 3.0]])
            my_dist = Uniform.from_logits(logits)
            my_dist.sample()
        """
        raise NotImplementedError

    @classmethod
    def get_partial_dist_cls(
        parent_cls: "Distribution", **partial_kwargs
    ) -> "Distribution":
        """Returns a partial child of TorchMultiActionDistribution.

        This is useful if inputs needed to instantiate the Distribution from logits
        are available, but the logits are not.
        """

        class DistributionPartial(parent_cls):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            @staticmethod
            def _merge_kwargs(**kwargs):
                """Checks if keys in kwargs don't clash with partial_kwargs."""
                overlap = set(kwargs) & set(partial_kwargs)
                if overlap:
                    raise ValueError(
                        f"Cannot override the following kwargs: {overlap}.\n"
                        f"This is because they were already set at the time this "
                        f"partial class was defined."
                    )
                merged_kwargs = {**partial_kwargs, **kwargs}
                return merged_kwargs

            @classmethod
            @override(parent_cls)
            def required_input_dim(cls, space: gym.Space, **kwargs) -> int:
                merged_kwargs = cls._merge_kwargs(**kwargs)
                assert space == merged_kwargs["space"]
                return parent_cls.required_input_dim(**merged_kwargs)

            @classmethod
            @override(parent_cls)
            def from_logits(
                cls,
                logits: TensorType,
                **kwargs,
            ) -> "DistributionPartial":
                merged_kwargs = cls._merge_kwargs(**kwargs)
                distribution = parent_cls.from_logits(logits, **merged_kwargs)
                # Replace the class of the returned distribution with this partial
                # This makes it so that we can use type() on this distribution and
                # get back the partial class.
                distribution.__class__ = cls
                return distribution

        # Substitute name of this partial class to match the original class.
        DistributionPartial.__name__ = f"{parent_cls}Partial"

        return DistributionPartial

    def to_deterministic(self) -> "Distribution":
        """Returns a deterministic equivalent for this distribution.

        Specifically, the deterministic equivalent for a Categorical distribution is a
        Deterministic distribution that selects the action with maximum logit value.
        Generally, the choice of the deterministic replacement is informed by
        established conventions.
        """
        raise NotImplementedError



from collections import defaultdict
from typing import Any, Callable


class LambdaDefaultDict(defaultdict):
    """A defaultdict that creates default values based on the associated key.

    Note that the standard defaultdict can only produce default values (via its factory)
    that are independent of the key under which they are stored.
    As opposed to that, the lambda functions used as factories for this
    `LambdaDefaultDict` class do accept a single argument: The missing key.
    If a missing key is accessed by the user, the provided lambda function is called
    with this missing key as its argument. The returned value is stored in the
    dictionary under that key and returned.

    Example:

        In this example, if you try to access a key that doesn't exist, it will call
        the lambda function, passing it the missing key. The function will return a
        string, which will be stored in the dictionary under that key.

        .. testcode::

            from ray.rllib.utils.lambda_defaultdict import LambdaDefaultDict

            default_dict = LambdaDefaultDict(lambda missing_key: f"Value for {missing_key}")
            print(default_dict["a"])

        .. testoutput::

            Value for a
    """  # noqa: E501

    def __init__(self, default_factory: Callable[[str], Any], *args, **kwargs):
        """Initializes a LambdaDefaultDict instance.

        Args:
            default_factory: The default factory callable, taking a string (key)
                and returning the default value to use for that key.
        """
        if not callable(default_factory):
            raise TypeError("First argument must be a Callable!")

        # We will handle the factory in __missing__ method.
        super().__init__(None, *args, **kwargs)

        self.default_factory = default_factory

    def __missing__(self, key):
        # Call default factory with the key as argument.
        self[key] = value = self.default_factory(key)
        return value
